from openpyxl.styles import PatternFill
from openpyxl import load_workbook
from OTCFinUtils.security import create_flow_headers, get_graph_token
from io import BytesIO
from http import HTTPStatus as status
import pandas as pd
import requests
import base64


# TODO - review and refactor code


def load_document(dataverse_url: str, token: str, document_path: str, drive: str, use_flow_headers: bool = False) -> BytesIO:
    """
    Load a document from SharePoint as a `BytesIO` object.
    """
    site_id, drive_id = get_connection_details(dataverse_url, token, drive, use_flow_headers)
    access_token = get_graph_token()

    headers = {
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }

    # TODO - review if provided logic corresponds to SharePoint URL format
    if document_path.startswith(drive):
        document_path = document_path[len(drive):]

    url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives/{drive_id}"
    url += f"/root:/{document_path}?select=id,@microsoft.graph.downloadUrl"

    response = requests.get(url, headers=headers)

    if response.status_code != status.OK.value:
        raise Exception(f"Failed to fetch JSON data. Status: {response.status_code}")

    json_data = response.json()
    download_url = json_data.get("@microsoft.graph.downloadUrl")

    if not download_url:
        raise Exception("Download URL not found in JSON data")

    download_response = requests.get(download_url)

    if download_response.status_code != status.OK.value:
        raise Exception(f"Failed to download file. Status: {download_response.status_code}")

    content = download_response.content

    return BytesIO(content)


# TODO - Remove dependence on Dataverse system variable, use key vault instead
# TODO - Use DVHandler object, instead of passing the url and token

def get_connection_details(dataverse_url: str, token: str, drive: str = "account", use_flow_headers: bool = False) -> tuple[str, str]:
    """
    Retrieves the SharePoint site id and drive id.
    Needed for reading and writing to SharePoint.
    """
    dataverse_url = f"{dataverse_url}/api/data/v9.2/new_systemvariableses?"
    dataverse_url += f"$select=new_value&"
    dataverse_url += f"$filter=%20new_name%20eq%20%27HTTP_SHAREPOINT_CONNECTION%27"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}

    response = requests.get(dataverse_url, headers=headers)
    
    if response.status_code != status.OK.value:
        raise Exception(f"Failed to get sharepoint connection details. Response: {response.text}")
    
    data = response.json()

    if "value" not in data or len(data["value"]) == 0:
        raise Exception("No data found in the response.")

    connection_details_url = data["value"][0]["new_value"]

    if use_flow_headers:
        flow_headers = create_flow_headers()
    else:
        flow_headers = None

    response = requests.post(connection_details_url, json={"drive": drive}, headers=flow_headers)

    if response.status_code != status.OK.value:
        raise Exception(f"Failed to get sharepoint connection details. Response: {response.text}")

    site_id = response.json().get("siteid")
    drive_id = response.json().get("driveid")

    return site_id, drive_id


def create_excel_document(
    df: pd.DataFrame,
    dataverse: str,
    token: str,
    drive: str,
    document_path: str,
    file_name: str,
    sheet: str,
) -> requests.Response:
    """
    Creates an Excel document on SharePoint.
    """
    if document_path.startswith(drive):
        document_path = document_path[len(drive):]
    
    access_token = get_graph_token()
    site_id, drive_id = get_connection_details(dataverse, token, drive,use_flow_headers=True)

    url = f"https://graph.microsoft.com/v1.0/sites"
    url += f"/{site_id}/drives/{drive_id}/root:"
    url += f"/{document_path}/{file_name}:/content"

    headers = {
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/octet-stream",
    }

    with BytesIO() as output:
        with pd.ExcelWriter(output, engine="openpyxl") as writer:
            df.to_excel(writer, sheet_name=sheet, index=False)
        output.seek(0)
        response = requests.put(url, output, headers=headers)

    if response.status_code not in (status.CREATED.value, status.OK.value):
        raise Exception(f"Failed to upload file: {response.status_code}")

    return response


def update_excel(excel_content: bytes, new_json_rows: list) -> str:
    """
    Work in progress...
    
    The goal was to have a function that can 
    append rows to an Excel file.
    """
    df = pd.read_excel(pd.ExcelFile(BytesIO(excel_content)))

    # If DataFrame is empty, create columns using keys from the first row of new_json_rows
    if df.empty:
        df = pd.DataFrame(columns=new_json_rows[0].keys())

    # Convert each JSON row into a DataFrame and concatenate them with the existing DataFrame
    for row_data in new_json_rows:
        row_df = pd.DataFrame([row_data])
        df = pd.concat([df, row_df], ignore_index=True)

    # Convert DataFrame to Excel format and return as bytes
    output = BytesIO()
    df.to_excel(output, index=False)

    # Load the workbook
    output.seek(0)
    workbook = load_workbook(output)
    sheet = workbook.active

    if sheet is None:
        raise RuntimeError("Value of 'sheet' cannot be None.")

    # Get the range of new rows in the sheet
    start_row = len(df) - len(new_json_rows) + 2
    end_row = len(df) + 2

    # Apply yellow fill only to the range of new rows
    for row in range(start_row, end_row):
        for column in range(1, len(df.columns) + 1):
            cell = sheet.cell(row=row, column=column)
            cell.fill = PatternFill(
                start_color="FFFF00", end_color="FFFF00", fill_type="solid"
            )

    # Adjust column widths to fit the text
    for column in sheet.columns:
        max_length = 0
        column = [cell for cell in column if cell.value is not None]
        if column:
            max_length = max(len(str(cell.value)) for cell in column)
            adjusted_width = (max_length + 2) * 1.2
            sheet.column_dimensions[column[0].column_letter].width = adjusted_width

    # Save the workbook to BytesIO
    output.seek(0)
    workbook.save(output)

    updated_excel_content = output.getvalue()

    # Encode Excel content to base64
    updated_excel_base64 = base64.b64encode(updated_excel_content).decode("utf-8")

    return updated_excel_base64



from itertools import product
import json
import uuid
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Optional, Any
from bisect import bisect_right
import pandas as pd
import pandas_market_calendars as mcal
from more_itertools import windowed
from sqlalchemy import (
    Connection,
    create_engine,
    MetaData,
    Table,
    select,
    insert,
    update,
    text,
    and_,
    or_,
)
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from sqlalchemy import select, and_, or_, text
from sqlalchemy.engine import Connection
from uuid import uuid4

from io import BytesIO
import base64
import requests
import copy

import re
import numpy as np
from scipy.stats import norm
from enum import Enum
import bisect

from dotenv import load_dotenv
import os
import math

import datetime as datetime
from datetime import datetime as dt

import pandas as pd

from dotenv import load_dotenv
from io import BytesIO



import pandas as pd


import os



from datetime import datetime, date


load_dotenv()
server = os.getenv("SQL_SERVER")
user = os.getenv("SQL_USER")
password = os.getenv("SQL_PASSWORD")
database = os.getenv("SQL_DATABASE")

class SQLHandler:
    def __init__(self, server: str, user: str, password: str, database: str):
        if not all([server, user, password, database]):
            raise ValueError("Database credentials not fully provided.")

        connection_string = (
            f"mssql+pymssql://{user}:{password}@{server}/{database}"
        )

        self.engine: Engine = create_engine(connection_string, future=True)
        self.metadata = MetaData()
        self.metadata.reflect(bind=self.engine)
    
    def upsert_bulk(self, data: list[dict], table, id_column: str) -> list[str]:
        if not data:
            return []

        valid_columns = {c.name for c in table.columns}
        # Columns that DB auto-fills, so omit them from INSERT unless explicitly provided.
        default_constraint_cols = {"new_createdon", "new_statecode"}
        skip_update_cols = {"new_createdby"}
        cleaned_data = []
        
        # Track which columns were actually provided in the input data to preserve selective updates.
        provided_columns = set()

        # Normalize and validate
        for record in data:
            record = record.copy()

            # Map entity_id field to the target id_column.
            if "entity_id" in record:
                record[id_column] = record.pop("entity_id")

            # Extract only valid columns from the record.
            row = {k: v for k, v in record.items() if k in valid_columns}
            
            # Track which columns were provided in this record.
            provided_columns.update(row.keys())
            
            # Generate GUID if ID is missing or empty.
            if id_column not in row or not row[id_column]:
                row[id_column] = str(uuid.uuid4())

            # Fill missing columns with None (except those DB manages with defaults).
            for col in valid_columns:
                if col not in row:
                    if col not in default_constraint_cols:
                        row[col] = None

            cleaned_data.append(row)

        # Extract IDs for return value.
        ids = [str(row[id_column]) for row in cleaned_data]
        # Include only columns that were provided or aren't DB-managed defaults.
        cols = [c for c in valid_columns if c not in default_constraint_cols or c in provided_columns]

        # Format column list for SQL statements.
        insert_cols = ", ".join(f"[{c}]" for c in cols)

        with Session(self.engine) as session:
            connection = session.connection()
            # Create temp table with same schema as target to stage bulk operations.
            temp_table = f"#TempUpsert_{uuid.uuid4().hex[:8]}"
            connection.execute(
                text(f"SELECT TOP 0 * INTO {temp_table} FROM [dbo].[{table.name}]")
            )

            # Bulk insert all records into temp table for efficient processing.
            # insert_stmt = text(f"""
            #     INSERT INTO {temp_table} ({insert_cols})
            #     VALUES ({", ".join(f":{c}" for c in cols)})
            # """)
            # connection.execute(insert_stmt, cleaned_data)
            self._bulk_insert_batched(
                connection=connection,
                table_name=temp_table,
                columns=cols,
                rows=cleaned_data,
            ) 

            # Filter to only columns that user provided and are updatable.
            update_columns = [
                c for c in provided_columns
                if c not in skip_update_cols and c != id_column and c in valid_columns
            ]
            # Update Bulk existing records with provided columns only (selective update pattern).
            if update_columns:
                set_clause = ", ".join(f"T.[{c}] = S.[{c}]" for c in update_columns)
                update_stmt = f"""
                    UPDATE T
                    SET {set_clause}
                    FROM [dbo].[{table.name}] AS T
                    INNER JOIN {temp_table} AS S
                    ON T.[{id_column}] = S.[{id_column}];
                """
                connection.execute(text(update_stmt))

            # Insert Bulk records that don't already exist.
            insert_stmt2 = f"""
                INSERT INTO [dbo].[{table.name}] ({insert_cols})
                SELECT {insert_cols}
                FROM {temp_table} AS S
                WHERE NOT EXISTS (
                    SELECT 1 FROM [dbo].[{table.name}] AS T
                    WHERE T.[{id_column}] = S.[{id_column}]
                );
            """
            connection.execute(text(insert_stmt2))
            # Cleanup temp table.
            connection.execute(text(f"DROP TABLE {temp_table}"))
            session.commit()

        return ids

    def get_entities(self, table, select_columns: str | None = None, filters: dict[str, object] | list[dict[str, object]] | None = None):
        with self.engine.begin() as conn:
            # Columns
            if select_columns:
                cols = []
                for name in (c.strip() for c in select_columns.split(",")):
                    if not name:
                        continue
                    if name not in table.c:
                        raise ValueError(f"Column '{name}' not found in table '{table.name}'")
                    cols.append(table.c[name])
            else:
                cols = list(table.c)

            # No filters → simple select
            if not filters:
                stmt = select(*cols)
                rows = conn.execute(stmt).mappings().all()
                return [{k: self.serialize_value(v) for k, v in row.items()} for row in rows]

            # Normalize to list
            if isinstance(filters, dict):
                filters = [filters]

            return self._execute_with_temp_table(conn, table, cols, filters)

    def _execute_with_temp_table(self, conn, table, cols, filters):

        temp_name = f"#temp_filter_{uuid.uuid4().hex}"

        filter_columns = set()
        for f in filters:
            filter_columns.update(f.keys())
        filter_columns = list(filter_columns)

        # Validate columns
        for col in filter_columns:
            if col not in table.c:
                raise ValueError(f"Filter column '{col}' not found in table '{table.name}'")

        # normalize filters into list of dicts with all combinations for multi-valued entries
        expanded_rows = []
        for group in filters:
            keys = []
            values = []
            for col in filter_columns:
                v = group.get(col, None)

                if isinstance(v, (list, tuple, set)):
                    vals = list(v)
                    if not vals:
                        vals = [None]
                    values.append(vals)
                else:
                    values.append([v])

                keys.append(col)

            for combo in product(*values):
                expanded_rows.append(dict(zip(keys, combo)))

        if not expanded_rows:
            return []
        
        # ----------------------------
        # Create temp table
        # ----------------------------
        column_defs = []
        for col_name in filter_columns:
            col_type = table.c[col_name].type.compile(self.engine.dialect)
            column_defs.append(f"[{col_name}] {col_type}")

        create_sql = f"""
            CREATE TABLE {temp_name} (
                {", ".join(column_defs)}
            )
        """
        conn.execute(text(create_sql))

        # ----------------------------
        # Bulk insert in temp table
        # ----------------------------
        # insert_sql = text(f"""
        #     INSERT INTO {temp_name} ({", ".join(f"[{c}]" for c in filter_columns)})
        #     VALUES ({", ".join(f":{c}" for c in filter_columns)})
        # """)
        # conn.execute(insert_sql, expanded_rows)
        self._bulk_insert_batched(
            connection=conn,
            table_name=temp_name,
            columns=filter_columns,
            rows=expanded_rows,
        ) 

        # ----------------------------
        # NULL-SAFE JOIN CONDITIONS 
        # ----------------------------
        join_conditions = " AND ".join(
            f"(t.[{c}] = f.[{c}] OR (t.[{c}] IS NULL AND f.[{c}] IS NULL))"
            for c in filter_columns
        )

        select_list = ", ".join(f"t.[{c.name}]" for c in cols)
        schema = table.schema or "dbo"

        query_sql = f"""
            SELECT {select_list}
            FROM [{schema}].[{table.name}] t
            INNER JOIN {temp_name} f
                ON {join_conditions}
        """

        rows = conn.execute(text(query_sql)).mappings().all()

        conn.execute(text(f"DROP TABLE {temp_name}"))

        return [{k: self.serialize_value(v) for k, v in row.items()} for row in rows]
    

    def _bulk_insert_batched(self, connection, table_name: str, columns: list[str], rows: list[dict],
        MAX_PARAMS: int = 2000): # stay below SQL Server 2100 limit
        """
        High-performance batched multi-row INSERT.
        """

        if not rows:
            return

        col_count = len(columns)

        # Calculate safe batch size dynamically
        batch_size = max(1, MAX_PARAMS // col_count)

        insert_cols = ", ".join(f"[{c}]" for c in columns)

        for i in range(0, len(rows), batch_size):
            batch = rows[i:i + batch_size]

            values_sql = []
            params = {}

            for row_index, row in enumerate(batch):
                row_params = []
                for col in columns:
                    param_name = f"{col}_{row_index}"
                    row_params.append(f":{param_name}")
                    params[param_name] = row.get(col)
                values_sql.append(f"({', '.join(row_params)})")

            stmt = text(f"""
                INSERT INTO {table_name} ({insert_cols})
                VALUES {', '.join(values_sql)}
            """)

            connection.execute(stmt, params)
    @staticmethod
    def serialize_value(value):
        if isinstance(value, (datetime, date)):
            return value.isoformat()
        elif isinstance(value, Decimal):
            return float(value)
        elif isinstance(value, uuid.UUID):
            return str(value)
        elif isinstance(value, bytes):
            return value.hex() 
        elif isinstance(value, bool):
            return bool(value) 
        elif value is None:
            return None
        elif isinstance(value, (list, dict)):
            return json.dumps(value)
        return value


from io import BytesIO
from OTCFinUtils.utils import (
    cast_data,
    choice_value,
    default_value,
    evaluate_condition,
    extract_sheets,
    extract_sorted_categories,
    get_skip_flag,
    group_mappings,
    is_float_value,
    is_valid,
    process_data_type, 
    extract_lookup_column, 
    extract_file_data, 
    is_date, 
    extract_date_string,
    tokenize,
    trim_string_fields,
    evaluate_formula_expression_generic,
    resolve_condition_output_generic
)
from OTCFinUtils.data_structs import (
    DEFAULT_SHEET_CHUNK_SIZE,
    NOT_DEFINED,
    Condition,
    Category,
    ResponseIds, 
    EntityLookups, 
    MappingFields, 
    IDParams,
    Lookup,
    KeyLookupParams,
    EntityKeyLookupParams,
)
from OTCFinUtils.mapper_loading import (
    get_buffer_ids,
    get_buffer_lookups,
)
from OTCFinUtils.dataverse_handler import DVHandler
from OTCFinUtils.sql_handler import SQLHandler
from OTCFinUtils.buffer import Buffer
from typing import Any, Optional
import pandas as pd
import json


class DataMapper:
    """
    Used for loading data into the Dataverse with the help of mappings from the File Mappings table.

    **IMPORTANT TO UNDERSTAND:**

    The code follows a somewhat specific logic to implement this.
    Without understanding the logic, it will be difficult to follow the code.

    On a high level, the code does the following steps:

    - Reads the provided file / data (input data)
    - Reads the provided mappings
    - Applies the mappings to the input data
    - Extracts query params for getting the entity ids (for checking if the data has already been loaded)
    - Extracts query params for getting the lookup ids for the input data
    - Queries the Dataverse for the entity ids (so we can know whether to perform a POST or PATCH request)
    - Queries the Dataverse for the lookup ids (so we can link the input data to the lookups)
    - Assigns the entity ids to the input data
    - Assigns the lookup ids to the input data
    - Performs a bulk upsert operation to the Dataverse
    - Returns the ids of the created and updated entities
    """
    
    def __init__(
        self,
        dvh: DVHandler,
        owner_id: str,
        user_id: str,
        sql_handler: Optional[SQLHandler] = None
    ) -> None:
        """
        Because we constantly interact with the DV API.
        Need to keep track of the connection information.
        """
        self.dvh: DVHandler = dvh
        self.owner_id: str = owner_id
        self.user_id: str = user_id
        # Optional SQL handler used when mapping to SQL instead of Dataverse
        self.sql_handler: Optional[SQLHandler] = sql_handler
        self.SHEET_CHUNK_SIZE: int = DEFAULT_SHEET_CHUNK_SIZE
        self.current_row: int = 0
        self.current_column: str = NOT_DEFINED
        self.current_sheet: str = NOT_DEFINED
        self.category_seen_keys: dict[Category, set[tuple]] = {}
        
        """
        Stored in order to query the table definitions only once,
        and not for every entity.
        """
        self.table_definitions:     dict[str, dict] = dict()

        """
        For storing buffered data that will be upserted in bulk.
        """
        self.buffer = Buffer()


    def _resolve_type_conflicts(self) -> None:
        """
        Necessary in the case the file types don't match the DV types.
        """
        for entity_data in self.buffer.category_data:
            entity_data = process_data_type(entity_data)
        return

    
    def _table_definition(self, table: str, fields: list[str]) -> dict:
        """
        Uses a caching approach to minimize number of queries to DV.
        """
        select_param = ",".join(fields)
        if table not in self.table_definitions:
            self.table_definitions[table] = self.dvh.get_table_definition(table, select_param)
        return self.table_definitions[table]


    def _resolve_empty_id_conditions(self) -> None:
        """
        In the case the ID parameter buffer has items with
        an empty list of conditions to filter by, add an 
        "impossible" condition.

        In such cases the not DV should return any entity ID.
        """
        for entity_id_params in self.buffer.category_id_params:
            if not entity_id_params.conditions:
                # An impossible condition, so the DV returns empty data
                entity_id_params.conditions = ["statecode eq 10"]
        return


    def _link_ids_with_data(self, ids: list[str]) -> None:
        """
        For adding the ids to the data buffer.
        Necessary for the bulk upsert operation.
        """
        for entity, id in zip(self.buffer.category_data, ids):
            entity["entity_id"] = id


    def _link_lookups(self, lookups: list[EntityLookups]) -> None:
        """
        For adding the lookups to the data buffer.
        Necessary for the bulk upsert operation.
        """
        if len(self.buffer.category_data) != len(lookups):
            raise RuntimeError(f"Category data buffers size ({len(self.buffer.category_data)}) and lookup list size ({len(lookups)}) do not match.")

        for entity, entity_lookups in zip(self.buffer.category_data, lookups):
            for lookup in entity_lookups:
                column = f"{lookup.bind_column}@odata.bind"
                if not lookup.id:
                    entity[column] = None
                else:
                    entity[column] = f"{lookup.table}({lookup.id})"

    
    def _upsert_data_buffer(self, category: Category) -> ResponseIds:
        """
        This operation is the ultimate goal of the whole code.
        Now accepts category to determine SQL vs Dataverse upsert per category.
        """
        try:
            # Determine upsert destination from category
            sql_table_name = category.sql_table

            if sql_table_name is not None:
                # SQL upsert path
                if not hasattr(self, 'sql_handler') or self.sql_handler is None:
                    raise RuntimeError(
                        f"SQL handler not initialized but category requires SQL upsert to {sql_table_name}"
                    )

                # Get SQLAlchemy table object
                if sql_table_name not in self.sql_handler.metadata.tables:
                    raise RuntimeError(
                        f"SQL table '{sql_table_name}' not found in database metadata"
                    )

                sql_table = self.sql_handler.metadata.tables[sql_table_name]

                # Format data for SQL (remove Dataverse-specific fields, apply SQL column renames)
                category_mappings = self.grouped_mappings.get(category, [])

                # Track which records had an existing ID (classified as 'updated')
                existing_id_flags = [
                    bool(entity.get("entity_id")) for entity in self.buffer.category_data
                ]

                self.format_data_for_sql_upsert(category_mappings)

                # Determine ID column (convention: table's primary ID column)
                # Using Dataverse table name + "id" as convention
                id_column = f"{category.table}id"

                sql_ids: list[str] = self.sql_handler.upsert_bulk(
                    self.buffer.category_data,
                    sql_table,
                    id_column
                )
                # Build ResponseIds for SQL upsert
                response = ResponseIds()
                for was_existing, entity_id in zip(existing_id_flags, sql_ids):
                    table_key = sql_table_name
                    if was_existing:
                        if table_key not in response.updated_entity_ids:
                            response.updated_entity_ids[table_key] = set()
                        response.updated_entity_ids[table_key].add(entity_id)
                    else:
                        if table_key not in response.created_entity_ids:
                            response.created_entity_ids[table_key] = set()
                        response.created_entity_ids[table_key].add(entity_id)
            else:
                # Dataverse upsert path (existing)
                response = self.dvh.upsert_bulk(self.buffer.category_data)

            return response

        except Exception as e:
            raise RuntimeError(f"Error while upserting the data: {e}")


    def format_data_for_sql_upsert(self, category_mappings: list[dict]) -> None:
        """
        Removes Dataverse-specific fields from buffer data before SQL upsert.

        Dataverse fields like @odata.bind, table_name, id_column need to be removed
        as they're not valid SQL columns.

        For lookup fields, extracts the GUID from @odata.bind and creates the corresponding
        foreign key field using the lowercase field name (e.g., new_Fund@odata.bind -> new_fund).

        Applies SQL column name overrides from new_sqlcolumnname mapping field:
        - If a mapping has new_sqlcolumnname set, the DV column is renamed to that SQL column name.
        - This allows e.g. DV column 'new_MainAccount' to be written to SQL column 'new_mainaccountname'.
        """
        import re

        # Build rename dict: {dv_column_name (lowercase) -> sql_column_name} for mappings that have new_sqlcolumnname set
        sql_column_renames: dict[str, str] = {}
        for mapping in category_mappings:
            dv_col = mapping.get(MappingFields.DV_COLUMN, "")
            sql_col = mapping.get(MappingFields.SQL_COLUMN_NAME, None)
            if dv_col and sql_col:
                sql_column_renames[dv_col.lower()] = sql_col

        cleaned_data = []

        for entity in self.buffer.category_data:
            cleaned_entity = {}

            for key, value in entity.items():
                # Handle Dataverse lookup relationship fields
                if key.endswith("@odata.bind"):
                    if value:
                        # Extract the field name (remove @odata.bind suffix, lowercase to match SQL column names)
                        field_name = key.replace("@odata.bind", "").lower()

                        # Extract GUID from value like "new_funds(guid-here)"
                        match = re.match(r'.*\(([a-f0-9-]+)\)', str(value))
                        if match:
                            guid = match.group(1)
                            # Apply SQL column rename if configured, else use lowercase field name
                            sql_col = sql_column_renames.get(field_name, field_name)
                            cleaned_entity[sql_col] = guid
                    # Don't include the @odata.bind field itself
                    continue

                elif key in ["table_name", "id_column", "skip"]:
                    # Skip metadata fields
                    continue

                else:
                    # Apply SQL column rename if configured, else keep key as-is
                    sql_col = sql_column_renames.get(key.lower(), key)
                    cleaned_entity[sql_col] = value

            cleaned_data.append(cleaned_entity)

        self.buffer.category_data = cleaned_data


    def _extract_lookup_vars(self, mapping: dict) -> tuple:
        """
        For increased readability, this logic is 
        grouped into one function.
        """
        try:
            table: str = mapping[MappingFields.LOOKUP_TABLE]
            relationship: str = mapping[MappingFields.LOOKUP_RELATIONSHIP]
            column = extract_lookup_column(mapping)
            
            table_definition_fields = ["PrimaryIdAttribute", "EntitySetName"]
            table_definition: dict = self._table_definition(table, table_definition_fields)
            
            id_column: str = table_definition.get("PrimaryIdAttribute", "")
            table_plural: str = table_definition.get("EntitySetName", "")

            return column, id_column, table_plural, relationship
        
        except KeyError as e:
            raise KeyError(f"Key error while extracting lookup variables: {e}")
        
    def _deduplicate_category_buffer_by_dv_keys(self, category: Category) -> None:
        """
        Remove duplicate entities from buffer.category_data based on
        the columns marked as Is Key = True for this category.
        Handles normal fields and lookup fields (@odata.bind).
        For SQL categories, uses SQL_COLUMN_NAME instead of DV_COLUMN.
        """
        is_sql_category = category.sql_table is not None
        source_column_field = MappingFields.DV_COLUMN if not is_sql_category else MappingFields.SQL_COLUMN_NAME
        
        key_columns: list[str] = []
        category_mappings = self.grouped_mappings[category]
        for mapping in category_mappings:
            if mapping[MappingFields.IS_KEY]:
                key_columns.append(mapping[source_column_field])

        if not key_columns:
            return

        def get_key_value(entity: dict, col: str):
            # For Dataverse we check normal field and @odata.bind variant
            if col in entity:
                return entity[col]
            # Only check @odata.bind for Dataverse category
            if not is_sql_category:
                bind_col = f"{col}@odata.bind"
                if bind_col in entity:
                    return entity[bind_col]
            return None

        seen_global = self.category_seen_keys.setdefault(category, set())

        unique_entities: list[dict] = []
        for entity in self.buffer.category_data:
            key = tuple(get_key_value(entity, col) for col in key_columns)
            if key in seen_global:
                continue
            seen_global.add(key)
            unique_entities.append(entity)

        self.buffer.category_data = unique_entities


    

    def _skip_entity_logic(self) -> None:
        """
        In the case we need to skip this entity (there is no entity in the file).
        """
        self.buffer.entity_data["skip"] = True

    
    def _lookup_logic(self, row: pd.Series, mapping: dict) -> None:
        """
        Extracts and appends the appropriate lookup parameters to the
        entity lookup buffer. 
        
        This buffer will later be loaded to the category buffer.
        
        The category buffer will be used for a bulk-reading of the 
        necessary lookup ids.

        If no lookup value is found in the file, 
        run the skip-entity logic.
        """
        column, id_column, table, relationship = self._extract_lookup_vars(mapping)        
        dv_column: str = mapping[MappingFields.DV_COLUMN]
        file_data = extract_file_data(row, mapping)

        if not file_data:
            self._skip_entity_logic()
            return

        if isinstance(file_data, pd.Timestamp):
            file_data = file_data.strftime("%Y-%m-%d")
            condition = f"{column} eq {file_data}"
        else:
            condition = f"{column} eq '{file_data}'"
        
        if not relationship in self.buffer.entity_lookup_params:
            self.buffer.entity_lookup_params[relationship] = IDParams(
                table=table,
                id_column=id_column,
                bind_column=dv_column
            )
        
        self.buffer.entity_lookup_params[relationship].conditions.append(condition)


    def _setup_key_parameters(self, loading_data: Any, table_column_name: str) -> None:
        """
        Loads the parameters for getting the id of the entity based on its key.
        """
        try:
            if is_date(loading_data):
                date = extract_date_string(loading_data)
                condition = f"{table_column_name} eq {date}"
            elif isinstance(loading_data, (int, float)):
                condition = f"{table_column_name} eq {loading_data}"
            else:
                condition = f"{table_column_name} eq '{loading_data}'"

            self.buffer.entity_id_params.conditions.append(condition)
        
        except Exception as e:
            raise RuntimeError(f"Error while running 'key parameter logic': {e}")


    def _key_logic(self, row: pd.Series, mapping: dict, SOURCE_TABLE_COLUMN_NAME: str) -> None:
        """
        Has two main steps:
        
        - Extracts and loads the appropriate id parameters in the
          entity id params buffer. This buffer will later be used
          to get the actual DV id of the entity, and not just the
          logical key of the entity.

        - Loads the logical key data to the entity data buffer.
          The entity data buffer will later be loaded to the
          category data buffer, which will be bulk-upserted to
          the DV.
        """
        try:
            file_column: str = mapping[MappingFields.FILE_COLUMN]
            file_data = row[file_column]
            table_column_name: str = mapping[SOURCE_TABLE_COLUMN_NAME]
            self._setup_key_parameters(file_data, table_column_name)
            self.buffer.entity_data[table_column_name] = file_data
        
        except Exception as e:
            raise RuntimeError(f"Error while running 'key logic': {e}")
        

    def _save_key_lookup_column(self, mapping: dict) -> None:
        """
        For getting the ids later, if some lookups are part of the key.
        """
        dv_column = mapping[MappingFields.DV_COLUMN]
        table = mapping[MappingFields.LOOKUP_TABLE]
        lookup_id_column = f"{table}id"
        lookup_id = None
        params = KeyLookupParams(lookup_id_column, lookup_id)
        self.buffer.entity_key_lookup_params[dv_column] = params

    
    def _key_lookup_logic(self, row: pd.Series, mapping: dict) -> None:
        """
        In the case the mapping is for a lookup and for a key of the entity.
        This is one of the most complicated cases. Later we get the id of the
        key using the 'expand' option in the DV API.

        We don't need to add the file data to the data buffer.
        This step is already done in the link-lookups step.
        """
        try:
            self._lookup_logic(row, mapping)
            self._save_key_lookup_column(mapping)
            
        except Exception as e:
            raise RuntimeError(f"Error while runnung 'key lookup' logic: {e}")


    def _choice_logic(self, row: pd.Series, mapping: dict) -> None:
        """
        In the case the mapping is for a choice. Assigns 
        the DV choice value in the entity buffer. If the
        file field is empty, does not add anything to the
        data buffer.
        """
        try:
            file_column: str = mapping[MappingFields.FILE_COLUMN]
            dv_column: str = mapping[MappingFields.DV_COLUMN]
            file_data = row[file_column]
            
            if not file_data or pd.isna(file_data):
                return
            
            if not isinstance(file_data, str):
                raise RuntimeError(f"File value must be a string: {file_data}")
            
            dv_choice_value = choice_value(file_data, mapping)
            self.buffer.entity_data[dv_column] = dv_choice_value

        except Exception as e:
            raise RuntimeError(f"Error in choice logic [{e}], for mapping: [{mapping}]")

    
    def _default_value_logic(self, mapping: dict, SOURCE_TABLE_COLUMN_NAME: str) -> Any:
        """
        In the case the mapping is for a default value.
        """
        try:
            table_column_name = mapping[SOURCE_TABLE_COLUMN_NAME]
            default_value_string = mapping[MappingFields.DEFAULT_VALUE]
            result = default_value(default_value_string, static_mappings=self.static_mappings)
            self.buffer.entity_data[table_column_name] = result
            return result
        except Exception as e:
            raise RuntimeError(f"Error in default-value logic: [{e}] for mapping: [{mapping}]")
        

    def _default_value_key_logic(self, mapping: dict, SOURCE_TABLE_COLUMN_NAME: str) -> None:
        """
        If the mapping defines a default value and if this field is part of the entity key.
        """
        default_value = self._default_value_logic(mapping, SOURCE_TABLE_COLUMN_NAME)
        table_column_name: str = mapping[SOURCE_TABLE_COLUMN_NAME]
        self._setup_key_parameters(default_value, table_column_name)
        
    def _lookup_default_value_logic(self, mapping: dict) -> None:
        """
        Lookup mapping that always uses the DEFAULT VALUE,
        completely ignoring any file column value.
        """
        column, id_column, table, relationship = self._extract_lookup_vars(mapping)
        dv_column: str = mapping[MappingFields.DV_COLUMN]

        default_value_string = mapping[MappingFields.DEFAULT_VALUE]
        if not default_value_string:
            self._skip_entity_logic()
            return

        lookup_value = default_value(default_value_string, getattr(self, "static_mappings", None))

        if not lookup_value:
            self._skip_entity_logic()
            return

        if isinstance(lookup_value, pd.Timestamp):
            lookup_value = lookup_value.strftime("%Y-%m-%d")
            condition = f"{column} eq {lookup_value}"
        else:
            condition = f"{column} eq '{lookup_value}'"

        if relationship not in self.buffer.entity_lookup_params:
            self.buffer.entity_lookup_params[relationship] = IDParams(
                table=table,
                id_column=id_column,
                bind_column=dv_column,
            )

        self.buffer.entity_lookup_params[relationship].conditions.append(condition)



    def _normal_data_logic(self, row: pd.Series, mapping: dict, SOURCE_TABLE_COLUMN_NAME: str) -> None:
        """
        The simplest case - just map the file data to the data buffer.
        No modifications to the data are necessary.
        """
        try:
            table_column_name = mapping[SOURCE_TABLE_COLUMN_NAME]
            file_column = mapping[MappingFields.FILE_COLUMN]
            self.buffer.entity_data[table_column_name] = row[file_column]
        except Exception as e:
            raise RuntimeError(f"Error in noraml-data-field logic: [{e}] for mapping [{mapping}]")
        

    def _condition_logic(self, row: pd.Series, mapping: dict) -> None:
        condition_string = mapping[MappingFields.CONDITION]
        if not condition_string:
            return

        # Parse condition JSON
        condition_dict = json.loads(condition_string)
        condition = Condition.create(condition_dict)

        # Decide true/false using the existing is_valid
        condition_result = is_valid(condition, row)

        if condition_result:
            raw_value = condition.result_value   # scalar or dict
        else:
            raw_value = condition.else_value

        # Turn raw_value into final DV value (supports FORMULA)
        final_value = self.resolve_condition_output(raw_value, row)

        dv_column = mapping[MappingFields.DV_COLUMN]

        # 1) set value on the entity (what we already had)
        self.buffer.entity_data[dv_column] = final_value

        # 2) 🔹 NEW: if this column is part of the key,
        #    also add it to the ID-lookup conditions
        is_key: bool = mapping[MappingFields.IS_KEY]
        if is_key:
            self._setup_key_parameters(final_value, dv_column)


        
    def evaluate_formula_expression(self, formula: str, row: pd.Series):
        return evaluate_formula_expression_generic(formula, row)

    def resolve_condition_output(self, value_def, row: pd.Series):
        return resolve_condition_output_generic(value_def, row)


    def _preprocess_row(self, row: pd.Series, mapping: dict) -> pd.Series:
        if mapping[MappingFields.DATA_TYPE] is None:
            return row
        
        data_type: str = mapping[MappingFields.DATA_TYPE].lower()
        file_column: str = mapping[MappingFields.FILE_COLUMN]
        file_value = row[file_column]
        
        if pd.notna(file_value) and isinstance(file_value, str):
            pass
        
        if data_type is not None and pd.notna(file_value):
            row[file_column] = cast_data(file_value, data_type)
        
        return row
    
    
    def _run_mapping_cycle(self, row: pd.Series, mapping: dict, category: Category) -> None:
        """
        Checks the current type of mapping and calls the corresponding 
        function for each type of mapping.
        """
        self.current_column = mapping[MappingFields.FILE_COLUMN]
        is_sql_category = category.sql_table is not None
        SOURCE_TABLE_COLUMN_NAME = MappingFields.DV_COLUMN if not is_sql_category else MappingFields.SQL_COLUMN_NAME

        try:
            is_lookup: bool = mapping[MappingFields.IS_LOOKUP]
            is_key: bool = mapping[MappingFields.IS_KEY]
            choice_dictionary: str = mapping[MappingFields.CHOICE_DICTIONARY]
            default_value_string: str = mapping[MappingFields.DEFAULT_VALUE]
            skip_entity: bool = get_skip_flag(row, mapping)
            
            row = self._preprocess_row(row, mapping)

            if evaluate_condition(row, mapping):
                self._condition_logic(row, mapping)

            elif choice_dictionary:
                self._choice_logic(row, mapping)
                
            elif is_lookup and default_value_string:
                self._lookup_default_value_logic(mapping)

            elif default_value_string and is_key:
                self._default_value_key_logic(mapping, SOURCE_TABLE_COLUMN_NAME)
            
            elif default_value_string:
                self._default_value_logic(mapping, SOURCE_TABLE_COLUMN_NAME)
            
            elif is_lookup and not is_key:
                self._lookup_logic(row, mapping)

            elif skip_entity:
                self._skip_entity_logic()
            
            elif is_key and not is_lookup:
                self._key_logic(row, mapping, SOURCE_TABLE_COLUMN_NAME)

            elif is_key and is_lookup:
                self._key_lookup_logic(row, mapping)
            
            else:
                self._normal_data_logic(row, mapping, SOURCE_TABLE_COLUMN_NAME)

        except Exception as e:
            raise e


    def _link_table_metadata(self, category: Category) -> None:
        """
        Because we need the metadata (the table name and id column)
        for correctly upserting the data.
        """
        try:
            if category.sql_table is None:
                table_definition_fields = ["PrimaryIdAttribute", "EntitySetName"]
                table_definition: dict = self._table_definition(category.table, table_definition_fields)
                
                id_column: str = table_definition.get("PrimaryIdAttribute", "")
                table_plural: str = table_definition.get("EntitySetName", "")

                self.buffer.entity_data["table_name"] = table_plural
                self.buffer.entity_data["id_column"] = id_column

                self.buffer.entity_id_params.table = table_plural
                self.buffer.entity_id_params.id_column = id_column
            else:
                self.buffer.entity_data["table_name"] = category.sql_table
                self.buffer.entity_data["id_column"] = f'{category.table}id'

                self.buffer.entity_id_params.table = category.sql_table
                self.buffer.entity_id_params.id_column = f'{category.table}id'
        
        except Exception as e:
            raise RuntimeError(f"Error while linking table metadata: {e}")


    def _run_entity_cycle(self, row: pd.Series, category: Category) -> None:
        """
        Resets all required buffers before everything else.
        
        For the entity of the specified category, in the given row, implements 
        all associated mappings. 
        
        Loads all associated entity buffer data into the category buffers. 
        
        Skips the loading steps if the mapping code has marked the entity as 
        a skip-entity.
        """
        self.current_row += 1

        try:
            self.buffer.reset_entity_buffers()
            category_mappings = self.grouped_mappings[category]

            for mapping in category_mappings:
                self._run_mapping_cycle(row, mapping, category)

            if "skip" in self.buffer.entity_data:
                return
            
            self._link_table_metadata(category)
            self.buffer.load_buffers()
        
        except Exception as e:
            raise e
        

    @staticmethod
    def _create_key_lookup_condition(entity_key_lookup_params: EntityKeyLookupParams, lookup: Lookup) -> str:
        """
        Helper function, for creating the DV API query filter parameter.
        """
        params = entity_key_lookup_params[lookup.bind_column]
        return f"{lookup.bind_column}/{params.lookup_id_column} eq '{lookup.id}'"
    
    
    def _attach_entity_lookups_to_buffer(self, entity_index: int, entity_lookups: EntityLookups) -> None:
        """
        So we can get the id of the entity, based on all its key-fields.
        """
        for lookup in entity_lookups:
            entity_params = self.buffer.category_key_lookup_params[entity_index]
            if lookup.bind_column not in entity_params:
                continue
            elif not lookup.id:
                raise RuntimeError(
                    f"Id of lookup '{lookup.bind_column}' cannot be null if part of key. "
                    f"Query conditions: {self.buffer.category_lookup_params[entity_index]}"
                )

            condition = DataMapper._create_key_lookup_condition(entity_params, lookup)
            self.buffer.category_id_params[entity_index].conditions.append(condition)
        

    def _attach_lookups_to_buffer(self, lookups: list[EntityLookups]) -> None:
        """
        Adds the filtering conditions for the key-
        lookup fields, so we can get the entity id.
        """
        for entity_index, entity_lookups in enumerate(lookups):
            self._attach_entity_lookups_to_buffer(entity_index, entity_lookups)


    def _process_lookups(self) -> None:
        """
        Because we need the lookup-fields to be 
        upserted as well.
        """
        lookups = get_buffer_lookups(self.dvh, self.buffer.category_lookup_params)
        self._link_lookups(lookups)
        self._attach_lookups_to_buffer(lookups)


    def _process_ids(self) -> None:
        """
        Because we need the ids to upserted as 
        well - for the entities that exist.
        """
        self._resolve_empty_id_conditions()
        entity_ids = get_buffer_ids(self.dvh, self.buffer.category_id_params)
        self._link_ids_with_data(entity_ids)


    def _process_ids_sql(self, category: Category) -> None:
        """
        Resolve existing entity IDs for SQL upsert categories using SQLHandler.

        Uses the key mappings (IS_KEY=True) for the category to build filters,
        queries SQL, and attaches 'entity_id' to each buffered entity when found.
        """
        sql_table_name = category.sql_table
        if not sql_table_name:
            return

        if not self.sql_handler:
            raise RuntimeError(
                f"SQL handler not initialized but category requires SQL upsert to {sql_table_name}"
            )

        if sql_table_name not in self.sql_handler.metadata.tables:
            raise RuntimeError(
                f"SQL table '{sql_table_name}' not found in database metadata"
            )

        sql_table = self.sql_handler.metadata.tables[sql_table_name]

        # Determine key mappings for this category
        category_mappings = self.grouped_mappings[category]
        key_mappings = [
            m for m in category_mappings if m[MappingFields.IS_KEY] is True
        ]
        if not key_mappings:
            return

        # Build list of SQL key column names (after applying SQL column override)
        sql_key_columns: list[str] = []
        for m in key_mappings:
            sql_override = m.get(MappingFields.SQL_COLUMN_NAME)
            if sql_override:
                sql_key_columns.append(sql_override)

        # Build filters list: each dict corresponds to one entity's key filter
        filters: list[dict[str, Any]] = []
        for entity in self.buffer.category_data:
            group: dict[str, Any] = {}
            for m, sql_col in zip(key_mappings, sql_key_columns):
                col_attr = m[MappingFields.SQL_COLUMN_NAME]
                group[sql_col] = entity.get(col_attr)
            filters.append(group)

        if not filters:
            return

        id_column = f"{category.table}id"
        select_columns = ", ".join([id_column] + sql_key_columns) if sql_key_columns and id_column else None
        rows = self.sql_handler.get_entities(
            sql_table,
            select_columns=select_columns,
            filters=filters
        )

        # Index SQL rows by key tuple for fast lookup
        def _row_key(row: dict) -> tuple:
            return tuple(row.get(col) for col in sql_key_columns)

        index: dict[tuple, Any] = {}
        for row in rows:
            index[_row_key(row)] = row.get(id_column)

        # Attach found IDs back to buffered entities
        for entity in self.buffer.category_data:
            key = tuple(
                entity.get(m[MappingFields.SQL_COLUMN_NAME]) for m in key_mappings
            )
            entity_id = index.get(key)
            if entity_id:
                entity["entity_id"] = entity_id


    def _process_buffered_data(self) -> None:
        """
        Because we need the lookups and ids to 
        be added to the loading data.
        """
        self._process_lookups()
        self._process_ids()


    def _link_data_owner(self) -> None:
        """
        Add the "global" account id as the owner 
        for all buffered entity data.
        """
        for entity in self.buffer.category_data:
            entity["ownerid@odata.bind"] = f"teams({self.owner_id})"

    
    def _postprocess_data(self, category: Category) -> None:
        """
        Because we need to add missing data and 
        execute clean ups before upserting.
        """
        self._resolve_type_conflicts()
        # Dataverse categories
        if category.sql_table is None:
            self._process_buffered_data()
            self._link_data_owner()
        # SQL categories
        else:
            self._process_ids_sql(category)


    def _run_category_cycle(self, chunk, category):
        try:
            for _, row in chunk.iterrows():
                self._run_entity_cycle(row, category)
        except Exception as e:
            raise Exception(
                f"Error: {e} | "
                f"Sheet: {self.current_sheet} | "
                f"Column: {self.current_column} | "
                f"Row: {self.current_row}"
            )

        self._postprocess_data(category)
        self._deduplicate_category_buffer_by_dv_keys(category)
        category_response_ids = self._upsert_data_buffer(category)
        self.buffer.reset_category_buffers()
        return category_response_ids
    

    def extract_category_key_columns(self, category: Category) -> list[str]:
        category_mappings = self.grouped_mappings[category]
        key_mappings = [mapping for mapping in category_mappings if mapping[MappingFields.IS_KEY] == True]
        key_file_columns = []
        for mapping in key_mappings:
            key_file_column = mapping[MappingFields.FILE_COLUMN]
            key_file_columns.append(key_file_column)
        return key_file_columns
    

    def extract_data_chunk(self, category_distinct_data: pd.DataFrame, index: int) -> pd.DataFrame:
        start_index = index * self.SHEET_CHUNK_SIZE
        end_index = start_index + self.SHEET_CHUNK_SIZE
        data_chunk = category_distinct_data.iloc[start_index:end_index]
        return data_chunk
    

    def load_category_data(self, category_distinct_data: pd.DataFrame, category: Category) -> ResponseIds:
        category_response_ids = ResponseIds()
        num_chunks = len(category_distinct_data) // self.SHEET_CHUNK_SIZE + 1
        
        for index in range(num_chunks):
            data_chunk = self.extract_data_chunk(category_distinct_data, index)
            chunk_response_ids = self._run_category_cycle(data_chunk, category)
            category_response_ids.merge(chunk_response_ids)
        
        return category_response_ids
    

    @staticmethod
    def extract_sheet_column_converters(column_converters: Optional[dict], sheet: str) -> Optional[dict]:
        if column_converters is None:
            return None
        
        return column_converters[sheet]
    
    
    def _load_sheet(
        self, 
        input_data: BytesIO | list[dict], 
        categories: list[Category], 
        sheet: str,
        header_row_number: int,
        column_type_converters: Optional[dict],
    ) -> ResponseIds:
        """
        Gets the sheet metadata and runs the loading 
        process for the given sheet. Returns the ids 
        of the updated and created entities.
        """
        # Reset the file row number for each sheet
        self.current_sheet = sheet
        self.current_column = NOT_DEFINED
        self.current_row = 0
        
        sheet_response_ids = ResponseIds()

        if isinstance(input_data, BytesIO):
            sheet_column_converters = DataMapper.extract_sheet_column_converters(column_type_converters, sheet)
            sheet_data = pd.read_excel(
                io=input_data, 
                sheet_name=sheet, 
                header=header_row_number, 
                keep_default_na=False, 
                na_values=[""],
                converters=sheet_column_converters,
            )
        elif isinstance(input_data, list):
            sheet_data = pd.DataFrame(input_data)

        sheet_data = trim_string_fields(sheet_data)
        
        for category in categories:
            category_key_columns = self.extract_category_key_columns(category)
            
            if len(category_key_columns) == 0:
                raise Exception(f"Entity key columns are not provided in mappings for: {category}")
            
            category_response_ids = self.load_category_data(sheet_data, category)
            sheet_response_ids.merge(category_response_ids)
        
        return sheet_response_ids


    # TODO - Add sheet name to parameter list - not needed anymore?
    # TODO - Remove table name and sheet name from categories - not needed anymore?

    def map_data(
        self, 
        input_data: BytesIO | list[dict], 
        mappings: list[dict], 
        sheet_chunk_size: Optional[int] = None,
        header_row_number: int = 0,
        column_converters: Optional[dict] = None,
        static_mappings: Optional[dict] = None,
    ) -> ResponseIds:
        """
        Prepare mappings, categories, and sheet names. Run the 
        mappings for each sheet. Return the ids of the updated 
        and created entities, grouped by table names.
        """
        # Override loading configuration parameter
        if sheet_chunk_size is not None:
            self.SHEET_CHUNK_SIZE = sheet_chunk_size
            
        self.static_mappings = static_mappings or {}
        self.grouped_mappings = group_mappings(mappings)
        categories = extract_sorted_categories(self.grouped_mappings)
        
        sheets: list[str] = extract_sheets(categories)
        final_response_ids = ResponseIds()

        for sheet in sheets:
            sheet_categories = list(filter(lambda x: x.sheet == sheet, categories))
            sheet_response_ids = self._load_sheet(input_data, sheet_categories, sheet, header_row_number, column_converters)
            final_response_ids.merge(sheet_response_ids)

        return final_response_ids



from OTCFinUtils.data_structs import (
    EntityKeyLookupParams, 
    EntityLookupParams, 
    IDParams,
)


class Buffer:
    def __init__(self) -> None:
        """
        Buffers for temporarily storing entity information.
        Need to use them for the batch reading / writing operations.
        """
        self.entity_data:                dict = dict()
        self.entity_id_params:           IDParams = IDParams()
        self.entity_lookup_params:       dict[str, IDParams] = dict()
        self.entity_key_lookup_params:   EntityKeyLookupParams = dict()

        """
        Buffers for temporarily storing information for all entities in a category.
        Need to use them for the batch reading / writing operations.
        """
        self.category_data:              list[dict] = []
        self.category_id_params:         list[IDParams] = []
        self.category_lookup_params:     list[EntityLookupParams] = []
        self.category_key_lookup_params: list[EntityKeyLookupParams] = []


    def _load_id_params_buffer(self) -> None:
        """
        Loads the buffer for the entity to the buffer for the category.
        This buffer is used for a batch query for the ids of the entities.
        Resets the entity buffer.
        """
        self.category_id_params.append(self.entity_id_params)
        self.entity_id_params = IDParams()


    def _load_data_buffer(self) -> None:
        """
        Loads the buffer for the entity to the buffer for the category.
        This is the buffer that will be bulk-upserted in the end.
        Resets the entity buffer.
        """
        self.category_data.append(self.entity_data)
        self.entity_data = dict()


    def _load_lookup_params_buffer(self) -> None:
        """
        Loads the lookup buffer for the entity, which is a dictionary,
        into the lookup buffer for the category. Transforms the dictionary 
        to a list, because that's how the the DV Handler requires it.
        Resets the entity buffer.
        """
        entity_lookup_params_list = list(self.entity_lookup_params.values())
        self.category_lookup_params.append(entity_lookup_params_list)
        self.entity_lookup_params = dict()


    def _load_key_lookup_params_buffer(self) -> None:
        """
        Loads the key-lookup params for the entity into
        the key-lookup buffer for the category of data.
        """
        self.category_key_lookup_params.append(self.entity_key_lookup_params)
        self.entity_key_lookup_params = dict()


    def reset_entity_buffers(self) -> None:
        """
        Resets all buffers for the current entity.
        Used in the entity cycle.
        """
        self.entity_data = dict()
        self.entity_id_params = IDParams()
        self.entity_lookup_params = dict()
        self.entity_key_lookup_params = dict()


    def reset_category_buffers(self):
        """
        Reset all category-level buffers.
        """
        self.category_id_params = []
        self.category_lookup_params = []
        self.category_key_lookup_params = []
        self.category_data = []


    def load_buffers(self) -> None:
        """
        Because we will query and upsert the data in batches,
        we need to store everything in our buffers.
        """
        try:
            self._load_lookup_params_buffer()
            self._load_id_params_buffer()
            self._load_data_buffer()
            self._load_key_lookup_params_buffer()
        
        except Exception as e:
            raise RuntimeError(f"Error while loading buffers: {e}")



"""
NOTE:
- This file is currently more of an integration script than a unit test.
- When executed as `python tests/test_dataverse_handler.py`, Python's import path
  starts in `tests/`, which can cause the *installed* `OTCFinUtils` package to be
  imported instead of the local source tree. That can lead to confusing runtime
  errors (e.g. missing methods).
"""

# TODO - write basic tests

# from pathlib import Path
# import sys
# from os import getenv

# # Ensure local repo sources are imported (repo_root/OTCFinUtils/...)
# REPO_ROOT = Path(__file__).resolve().parents[1]
# sys.path.insert(0, str(REPO_ROOT))

# from dotenv import load_dotenv
# from OTCFinUtils.dataverse_handler import DVHandler

# load_dotenv()

# dataverse =


# dvh = DVHandler(dataverse)



import unittest
from unittest.mock import patch, Mock, call
import unittest.test
from OTCFinUtils.data_structs import MappingFields, IDParams, Lookup
from OTCFinUtils.dataverse_handler import set_up_dv_handler, DVHandler
from OTCFinUtils.mapper import DataMapper

# TODO - add more fields and mappings - choice, condition, alternative column, exclude value, data type

mock_mappings = [
    # Test Column
    {
        MappingFields.DV_COLUMN: "new_testcolumn",
        MappingFields.DV_TABLE: "new_testtable",
        MappingFields.FILE_COLUMN: "Test Column",
        MappingFields.SHEET: "sheet 1",
        MappingFields.MAP_GROUP_ORDER: 1,
        MappingFields.LOOKUP_TABLE: None,
        MappingFields.LOOKUP_RELATIONSHIP: None,
        MappingFields.LOOKUP_COLUMN: None,
        MappingFields.DEFAULT_VALUE: None,
        MappingFields.DATA_TYPE: None,
        MappingFields.IS_LOOKUP: False,
        MappingFields.IS_KEY: False,
        MappingFields.IS_CHOICE: False,
        MappingFields.CHOICE_DICTIONARY: None,
        MappingFields.CONDITION: None,
        MappingFields.ALTERNATIVE_FILE_COLUMN: None,
        MappingFields.EXCLUDE_VALUE: None,
    },
    # Test Key Column
    {
        MappingFields.DV_COLUMN: "new_testkeycolumn",
        MappingFields.DV_TABLE: "new_testtable",
        MappingFields.FILE_COLUMN: "Test Key Column",
        MappingFields.SHEET: "sheet 1",
        MappingFields.MAP_GROUP_ORDER: 1,
        MappingFields.IS_KEY: True,
        MappingFields.LOOKUP_TABLE: None,
        MappingFields.LOOKUP_RELATIONSHIP: None,
        MappingFields.LOOKUP_COLUMN: None,
        MappingFields.DEFAULT_VALUE: None,
        MappingFields.DATA_TYPE: None,
        MappingFields.IS_LOOKUP: False,
        MappingFields.IS_CHOICE: False,
        MappingFields.CHOICE_DICTIONARY: None,
        MappingFields.CONDITION: None,
        MappingFields.ALTERNATIVE_FILE_COLUMN: None,
        MappingFields.EXCLUDE_VALUE: None,
    },
    # Test Lookup Column 1
    {
        MappingFields.DV_COLUMN: "new_TestLookupColumn",
        MappingFields.DV_TABLE: "new_testtable",
        MappingFields.FILE_COLUMN: "Test Lookup Column",
        MappingFields.SHEET: "sheet 1",
        MappingFields.IS_LOOKUP: True,
        MappingFields.IS_KEY: True,
        MappingFields.MAP_GROUP_ORDER: 1,
        MappingFields.LOOKUP_TABLE: "new_testtablelookup",
        MappingFields.LOOKUP_RELATIONSHIP: "lookup 1",
        MappingFields.LOOKUP_COLUMN: "[\"new_name\"]",
        MappingFields.DEFAULT_VALUE: None,
        MappingFields.DATA_TYPE: None,
        MappingFields.IS_CHOICE: False,
        MappingFields.CHOICE_DICTIONARY: None,
        MappingFields.CONDITION: None,
        MappingFields.ALTERNATIVE_FILE_COLUMN: None,
        MappingFields.EXCLUDE_VALUE: None,
    },
    # Test Lookup Column 2
    {
        MappingFields.DV_COLUMN: "new_TestLookupColumn2",
        MappingFields.DV_TABLE: "new_testtable",
        MappingFields.FILE_COLUMN: "Test Lookup Column 2",
        MappingFields.SHEET: "sheet 1",
        MappingFields.IS_LOOKUP: True,
        MappingFields.MAP_GROUP_ORDER: 1,
        MappingFields.LOOKUP_TABLE: "new_testtablelookup2",
        MappingFields.LOOKUP_RELATIONSHIP: "lookup 2",
        MappingFields.LOOKUP_COLUMN: "[\"new_name\"]",
        MappingFields.DEFAULT_VALUE: None,
        MappingFields.DATA_TYPE: None,
        MappingFields.IS_KEY: False,
        MappingFields.IS_CHOICE: False,
        MappingFields.CHOICE_DICTIONARY: None,
        MappingFields.CONDITION: None,
        MappingFields.ALTERNATIVE_FILE_COLUMN: None,
        MappingFields.EXCLUDE_VALUE: None,
    }
]

mock_data = [
    {
        "Test Column": "TEST VALUE",
        "Test Key Column": "TEST VALUE",
        "Test Lookup Column": "TEST VALUE",
        "Test Lookup Column 2": "TEST VALUE",
    }
]

upsert_data_params = [
    {
        'new_testcolumn': 'TEST VALUE', 
        'new_testkeycolumn': 'TEST VALUE', 
        'table_name': 'new_testtables', 
        'id_column': 'new_testtableid', 
        'new_TestLookupColumn@odata.bind': 
        'new_testtablelookups(6a709716-2687-ef11-ac20-7c1e52194aa8)', 
        'new_TestLookupColumn2@odata.bind': 
        'new_testtablelookup2s(ab755a1d-2687-ef11-ac20-7c1e52194aa8)', 
        'entity_id': None, 
        'ownerid@odata.bind': 
        'teams(6746bba8-de61-ee11-8def-000d3a17af6f)'
    }
]

get_buffer_ids_params = [
    IDParams(
        table='new_testtables', 
        id_column='new_testtableid', 
        conditions=[
            "new_testkeycolumn eq 'TEST VALUE'", 
            "new_TestLookupColumn/new_testtablelookupid eq '6a709716-2687-ef11-ac20-7c1e52194aa8'"
        ], 
        bind_column=None,
    ),
]

get_buffer_ids_return_value = [None]

get_buffer_lookups_params = [[
    IDParams(
        table='new_testtablelookups', 
        id_column='new_testtablelookupid', 
        conditions=["new_name eq 'TEST VALUE'"], 
        bind_column='new_TestLookupColumn',
    ), 
    IDParams(
        table='new_testtablelookup2s', 
        id_column='new_testtablelookup2id', 
        conditions=["new_name eq 'TEST VALUE'"], 
        bind_column='new_TestLookupColumn2',
    ),
]]

get_buffer_lookups_return_value = [[
    Lookup(
        bind_column='new_TestLookupColumn', 
        table='new_testtablelookups', 
        id='6a709716-2687-ef11-ac20-7c1e52194aa8',
    ), 
    Lookup(
        bind_column='new_TestLookupColumn2', 
        table='new_testtablelookup2s', 
        id='ab755a1d-2687-ef11-ac20-7c1e52194aa8',
    ),
]]

table_definition_calls = [
    call("new_testtablelookup", "PrimaryIdAttribute,EntitySetName"),
    call("new_testtablelookup2", "PrimaryIdAttribute,EntitySetName"),
    call("new_testtable", "PrimaryIdAttribute,EntitySetName"),
]

table_definition_outputs = [
    {'PrimaryIdAttribute': 'new_testtablelookupid', 'EntitySetName': 'new_testtablelookups',},
    {'PrimaryIdAttribute': 'new_testtablelookup2id', 'EntitySetName': 'new_testtablelookup2s',},
    {'PrimaryIdAttribute': 'new_testtableid', 'EntitySetName': 'new_testtables',},
]

OTC_TEAM_ID = "6746bba8-de61-ee11-8def-000d3a17af6f"
DEV_DATAVERSE_URL = "https://org873d3f04.crm.dynamics.com/"

class TestMyClass(unittest.TestCase):
    def setUp(self):
        dataverse_url = DEV_DATAVERSE_URL
        owner_id = OTC_TEAM_ID
        user_id = "nonexistent-user-id"
        self.dvh = set_up_dv_handler(dataverse_url)
        self.mapper = DataMapper(self.dvh, owner_id, user_id)

    
    @patch("OTCFinUtils.mapper.get_buffer_lookups")
    @patch("OTCFinUtils.mapper.get_buffer_ids")
    @patch.object(DVHandler, "upsert_bulk")
    @patch.object(DVHandler, "get_table_definition")
    def test_run_logic_calls_upsert_bulk(
        self, 
        mock_get_table_definition: Mock,
        mock_upsert_bulk: Mock, 
        mock_get_buffer_ids: Mock, 
        mock_get_buffer_lookups: Mock
    ):
        # Configure the mock return values
        mock_get_table_definition.side_effect = table_definition_outputs
        mock_get_buffer_ids.return_value = get_buffer_ids_return_value
        mock_get_buffer_lookups.return_value = get_buffer_lookups_return_value

        # Call the tested function 
        self.mapper.map_data(mock_data, mock_mappings)

        # Assert that mocked functions were called with the correct data
        self.assertEqual(mock_get_table_definition.mock_calls, table_definition_calls)
        mock_get_buffer_lookups.assert_called_once_with(self.dvh, get_buffer_lookups_params)
        mock_get_buffer_ids.assert_called_once_with(self.dvh, get_buffer_ids_params)
        mock_upsert_bulk.assert_called_once_with(upsert_data_params)


if __name__ == '__main__':
    unittest.main()


import numbers
import re
import math
import json
import pandas as pd
from datetime import datetime
from typing import Optional, Any, Union
from OTCFinUtils.data_structs import _TOKEN_REGEX, MappingFields, Condition, Operator, Category



def chunked_list(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]

def extract_segment_object_line(lines: list[str]) -> str:
    for line in lines:
        if "@odata.context" in line:
            return line
    raise ValueError(f"Object line not found in batch response. Lines: {lines}")


def extract_segment_response_id(lines: list[str], id_column: str) -> Optional[str]:
    """
    From a list of lines of a batch response, returns 
    the id of the entity in the response. If a null
    id column name is provided, return None.
    """
    try:
        if not id_column:
            return None

        object_line = extract_segment_object_line(lines)
        json_object = json.loads(object_line)
        object_id: str = json_object[id_column]

        return object_id
    
    except json.JSONDecodeError as e:
        raise RuntimeError(f"JSON error in response id logic: {e}")
    except KeyError as e:
        raise RuntimeError(f"Key error in response id logic: {e}")
    except Exception as e:
        raise RuntimeError(f"Error in response id logic: {e}")


def extract_segment_status_code(lines: list[str]) -> int:
    """
    From a list of lines of a batch response,
    returns the HTTP status code.
    """
    status_code_line = lines[4]
    status_code_line_parts = status_code_line.split(" ")
    status_code = status_code_line_parts[1]
    status_code = int(status_code)
    
    return status_code


def extract_segment_table_name(lines: list[str]) -> str:
    """
    From a list of lines of a batch response,
    returns the table name of the entity using regex.
    """
    object_line = extract_segment_object_line(lines)
    pattern = r'metadata#([a-z0-9_]+)'
    pattern_matches = re.findall(pattern, object_line)
    
    if not pattern_matches:
        raise RuntimeError(f"Could not find table pattern in string: {object_line}")
    
    table = pattern_matches[0]
    return table


def process_data_type(data: dict) -> None:
    """
    Resolves conflicts between file data types and DV data types.
    """
    try:
        for key, value in data.items():
            if isinstance(value, pd.Timestamp):
                if pd.isna(value):
                    data[key] = None
                else:
                    data[key] = value.strftime("%Y-%m-%d")

            elif isinstance(value, datetime):
                data[key] = value.strftime("%Y-%m-%d")

            elif isinstance(value, str):
                # Try parsing string to date if the key is likely a date field
                if "date" in key.lower() and not re.search(r'\b\d{4}-\d{2}-\d{2}->\d{4}-\d{2}-\d{2}\b', value):
                    try:
                        parsed = pd.to_datetime(value)
                        data[key] = parsed.strftime("%Y-%m-%d")
                    except Exception:
                        data[key] = None

            elif value is None or (not isinstance(value, str) and math.isnan(value)):
                data[key] = None
    
    except Exception as e:
        raise RuntimeError(f"Error: {e}, while processing data object: {data}")
    

def extract_lookup_column(mapping: dict) -> str:
    """
    Gets the lookup column value from the 
    JSON string in the mapping.
    """
    try:
        column_string: str = mapping[MappingFields.LOOKUP_COLUMN]
        # TODO: Remove when no longer needed
        column_string = column_string.replace("'", '"')
        column = json.loads(column_string)[0]
        return column
    
    except Exception as e:
        raise RuntimeError(f"Error while extracting the lookup column: {e}")
    

def extract_file_data(row: pd.Series, mapping: dict):
    """
    Get the data from the appropriate file column, 
    depending on whether the data matches the "exclude 
    values" list from the mapping.
    """
    try:
        if data_equals_exclude_value(row, mapping):
            alternative_file_column: str = mapping[MappingFields.ALTERNATIVE_FILE_COLUMN]
            file_data = row[alternative_file_column]
        else:
            file_column: str = mapping[MappingFields.FILE_COLUMN]
            file_data = row[file_column]
        
        return trim_string_values(file_data)
    
    except Exception as e:
        raise RuntimeError(f"Could not extract file data. Error: {e}")
    

def data_equals_exclude_value(row: pd.Series, mapping: dict) -> bool:
    """
    Checks if the file data matches one of the 
    "exclude values" from the mapping.
    """
    try:
        file_column: str = mapping[MappingFields.FILE_COLUMN]
        exclude_values_list: str = mapping[MappingFields.EXCLUDE_VALUE]
        
        if not exclude_values_list:
            return False
        
        exclude_values: list = json.loads(exclude_values_list)
        
        if row[file_column] in exclude_values:
            return True
        
        return False
    
    except json.JSONDecodeError as e:
        raise ValueError(f"Could not decode JSON string: {exclude_values_list}. Error: {e}") # type: ignore
    
    except KeyError as e:
        raise KeyError(f"Key error for 'exclude value' logic: {e}")

    except Exception as e:
        raise RuntimeError(f"Error for 'exclude value' logic: {e}")
    

def trim_string_values(value: Any) -> Any:
    """
    If the value is a string trims and returns it.
    Returns all other types of values.
    """
    if isinstance(value, str):
        return value.strip()
    return value


def is_date(value: Any) -> bool:
    """
    Helper function.
    """
    if isinstance(value, pd.Timestamp):
        return True 
    
    elif isinstance(value, str) and re.fullmatch(r"^\d{4}-\d{2}-\d{2}$", value):
        return True
    
    return False


def extract_date_string(value: pd.Timestamp | str) -> str:
    """
    Helper function.
    """
    if isinstance(value, pd.Timestamp):
        return value.strftime("%Y-%m-%d")
    elif isinstance(value, str):
        return value
    else:
        raise RuntimeError(f"The provided value is not a Timestamp or string: {value}")


def choice_value(file_data: str, mapping: dict) -> int | bool:
    """
    Should return the DV value (number) of the choice we need,
    based on the file data value. 
    """
    choice_string: str = mapping[MappingFields.CHOICE_DICTIONARY]
    
    try:
        choice_mappings: list[dict] = json.loads(choice_string)
    except json.JSONDecodeError:
        raise RuntimeError(f"The choice dictionary: {choice_string} is not a valid JSON object.")
    
    choice_mapping: dict = normalize_choice_map(choice_mappings)
    
    # To normalize the choice strings
    file_data = str(file_data).upper()
    
    return choice_mapping[file_data]


def normalize_choice_map(choice_mappings: list[dict]) -> dict:
    """
    Should return a normalized dictionary of the mappings,
    from the un-normalized list of mappings.
    """
    try:
        choice_mapping = dict()
        
        for item in choice_mappings:
            dv_value: int | bool = item["Value"]
            excel_value: str = item["Excel_Value"]
            
            # To normalize the choice strings
            excel_value = excel_value.upper()
            
            choice_mapping[excel_value] = dv_value
        
        return choice_mapping
    
    except KeyError as e:
        raise KeyError(f"Key Error for choice map normalization: {e}")
    
    except Exception as e:
        raise RuntimeError(f"Error while normalizing choice map: {e}")
    

def default_value(default_value_string: str, column_value: Optional[str] = None, static_mappings: Optional[dict] = None):
    static_mappings = static_mappings or {}

    try:
        raw = default_value_string.strip()

        if (raw.startswith('"') and raw.endswith('"')) or (raw.startswith("'") and raw.endswith("'")):
            raw = raw[1:-1]

        default_value_info = json.loads(raw)

        dv_type = default_value_info.get("type", "")
        dv_value = default_value_info.get("value")
        dv_format = default_value_info.get("format", "")
        dv_old_format = default_value_info.get("old_format", "")

        if dv_type == "current_date":
            date_format = dv_format or "%Y-%m-%d"
            current_time = datetime.now()
            date_string = current_time.strftime(date_format)
            return pd.to_datetime(date_string)
        
        elif dv_type == "timestamp":
            parsed_time = datetime.strptime(column_value, dv_old_format) if column_value else None
            return parsed_time.strftime(dv_format) if parsed_time else None

        elif dv_type == "static_mapping":
            key = dv_value
            if key not in static_mappings:
                raise KeyError(f"Static mapping '{key}' not found")
            v = static_mappings[key]
            return v() if callable(v) else v

        else:
            return dv_value

    except json.JSONDecodeError:
        raise ValueError(f"Could not decode the string: {default_value_string}")

    except KeyError as e:
        raise KeyError(f"Could not read key for default value logic: {e}")

    
def is_float_value(val) -> bool:
    if isinstance(val, numbers.Real) and not isinstance(val, bool):
        return True
    if isinstance(val, str):
        try:
            float(val)
            return True
        except ValueError:
            return False
    return False


def tokenize(formula: str):
    tokens = []
    for mo in re.finditer(_TOKEN_REGEX, formula):
        kind = mo.lastgroup
        value = mo.group()
        if kind in ('COL', 'NUMBER', 'IDENT', 'STRING'):
            tokens.append((kind, value))
        elif kind == 'COMP':
            tokens.append(('COMP', value))
        elif kind == 'COMMA':
            tokens.append(('COMMA', value))
        elif kind == 'OP':
            tokens.append((value, value))  # '+', '-', '*', '/', '(', ')'
        elif kind == 'SKIP':
            continue
        elif kind == 'MISMATCH':
            raise ValueError(f"Unexpected character {value!r} in formula")
    return tokens


def cast_data(value: Any, data_type: str) -> Any:
    """
    Casts the provided value to the given data type.
    Return the value if the data type is unknown.
    """
    try:
        if pd.isna(value):
            return value
        elif data_type == "string":
            return str(value)
        elif data_type == "integer":
            return int(value)
        elif data_type == "decimal":
            return float(value)
        elif data_type == "boolean":
            return bool(value)
        else:
            return value
    
    except Exception as e:
        raise RuntimeError(f"Error while type-casting data: {e}")
    

def get_skip_flag(row: pd.Series, mapping: dict) -> bool:
    """
    Return a flag for whether to skip the current entity or not,
    depending on whether the file data matches one of the "exclude 
    values" in the mapping or if the file data is an empty value.
    """
    is_key: bool = mapping[MappingFields.IS_KEY]

    return (
        (is_key and data_is_empty(row, mapping)) or 
        data_equals_exclude_value(row, mapping)
    )


def data_is_empty(row: pd.Series, mapping: dict) -> bool:
    """
    Checks if the file data is an empty cell. Returns True 
    if empty. Returns False otherwise. Does not take into 
    consideration the alternative file column.
    """
    file_column: str = mapping[MappingFields.FILE_COLUMN]
    file_data = row[file_column]
    
    if pd.isna(file_data) or not file_data:
        return True
    
    return False


def evaluate_condition(row: pd.Series, mappping: dict) -> bool:
    """
    Checks if the mapping has a condition and 
    if the row satisfies the condtition. If true,
    returns the resulting value, otherwise "None".
    """
    try:
        condition_string = mappping[MappingFields.CONDITION]

        if not condition_string:
            return False
        
        condition_dict = json.loads(condition_string)
        condition = Condition.create(condition_dict)
        
        return is_valid(condition, row)
    
    except Exception as e:
        raise RuntimeError(f"Could not evaluate condition. Mapping: {mappping}. Error: {e}")
    

def is_valid(condition: Condition, row: pd.Series) -> bool:
    """
    Evaluates if the row satisfies the given condition.
    """
    if (
        condition.operator == Operator.EQUALS and 
        condition.is_static_value and 
        row[condition.column] == condition.compare_value
    ):
        return True
    
    elif (
        condition.operator == Operator.EQUALS and 
        not condition.is_static_value and 
        row[condition.column] == row[condition.compare_value]
    ):
        return True
    
    # ----------------------------------------------------
    
    elif (
        condition.operator == Operator.NOT_EQUALS and
        condition.is_static_value and
        row[condition.column] != condition.compare_value
    ):
        return True
    
    elif (
        condition.operator == Operator.NOT_EQUALS and
        not condition.is_static_value and
        row[condition.column] != row[condition.compare_value]
    ):
        return True
    
    # ----------------------------------------------------
    
    elif (
        condition.operator == Operator.IS_NULL and
        row[condition.column] == None
    ):
        return True
    
    elif (
        condition.operator == Operator.IS_NOT_NULL and
        row[condition.column] != None
    ):
        return True
    
    # ----------------------------------------------------
    
    elif (
        condition.operator == Operator.SMALLER and
        condition.is_static_value and
        row[condition.column] < condition.compare_value
    ):
        return True
    
    elif (
        condition.operator == Operator.SMALLER and
        not condition.is_static_value and
        row[condition.column] < row[condition.compare_value]
    ):
        return True
    
    # ----------------------------------------------------
    
    elif (
        condition.operator == Operator.GREATER and
        condition.is_static_value and
        row[condition.column] > condition.compare_value
    ):
        return True
    
    elif (
        condition.operator == Operator.GREATER and
        not condition.is_static_value and
        row[condition.column] > row[condition.compare_value]
    ):
        return True
    
    # ----------------------------------------------------

    elif (
        condition.operator == Operator.SMALLER_OR_EQUAL and
        condition.is_static_value and
        row[condition.column] <= condition.compare_value
    ):
        return True
    
    elif (
        condition.operator == Operator.SMALLER_OR_EQUAL and
        not condition.is_static_value and
        row[condition.column] <= row[condition.compare_value]
    ):
        return True
    
    # ----------------------------------------------------
    
    elif (
        condition.operator == Operator.GREATER_OR_EQUAL and
        condition.is_static_value and
        row[condition.column] >= condition.compare_value
    ):
        return True
    
    elif (
        condition.operator == Operator.GREATER_OR_EQUAL and
        not condition.is_static_value and
        row[condition.column] >= row[condition.compare_value]
    ):
        return True
    
    return False


def group_mappings(mappings: list[dict]) -> dict[Category, list[dict]]:
    """
    From all the mapping, return a dictionary of grouped mapping lists,
    based on their categories (sheet name + order number + DV table + SQL table).

    Validates that all mappings within a category have consistent SQL configuration.
    """
    grouped_mappings: dict[Category, list[dict]] = dict()

    for mapping in mappings:
        group: int = mapping[MappingFields.MAP_GROUP_ORDER]
        table: str = mapping[MappingFields.DV_TABLE]
        sheet: str = mapping[MappingFields.SHEET]

        # Extract SQL configuration (normalize None to False)
        upsert_to_sql: bool = mapping.get(MappingFields.UPSERT_TO_SQL, False) or False
        sql_table_name: str | None = mapping.get(MappingFields.SQL_TABLE_NAME, None) if upsert_to_sql else None

        # Create category with SQL table info
        file_mapping_category = Category(group, table, sheet, sql_table_name)

        if file_mapping_category not in grouped_mappings:
            grouped_mappings[file_mapping_category] = list()
        else:
            # Validate this mapping has same SQL config as others in category
            existing_mappings = grouped_mappings[file_mapping_category]
            first_mapping = existing_mappings[0]
            # Normalize None to False for comparison
            first_upsert_to_sql = first_mapping.get(MappingFields.UPSERT_TO_SQL, False) or False
            first_sql_table = first_mapping.get(MappingFields.SQL_TABLE_NAME, None)

            if upsert_to_sql != first_upsert_to_sql or sql_table_name != first_sql_table:
                raise ValueError(
                    f"Inconsistent SQL configuration in category (group={group}, table={table}, sheet={sheet}). "
                    f"All mappings in a category must have the same SQL upsert settings. "
                    f"First mapping: upsert={first_upsert_to_sql}, table={first_sql_table}. "
                    f"Current mapping: upsert={upsert_to_sql}, table={sql_table_name}."
                )

        grouped_mappings[file_mapping_category].append(mapping)

    return grouped_mappings


RowLike = Union[pd.Series, dict]

def evaluate_formula_expression_generic(formula: str, row: RowLike) -> Any:
    """
    Shared formula evaluator for importer/exporter.

    Row: pd.Series or dict.
    Formula uses [Column Name] tokens, e.g.:
      [Comm /K] * -1000
      [AcctCode] + '_' + [Symbol] + '_' + DATE([Trade Date]) + '_' + [Buys or Sells]
    """
    # normalize row to plain dict
    if isinstance(row, pd.Series):
        context = row.to_dict()
    elif isinstance(row, dict):
        context = row
    else:
        raise TypeError(f"Row must be dict or Series, got {type(row)}")

    tokens = tokenize(formula)
    pos = 0

    def peek():
        nonlocal pos
        return tokens[pos] if pos < len(tokens) else None

    def advance():
        nonlocal pos
        tok = peek()
        pos += 1
        return tok

    def apply_op(op, left, right):
        # text concatenation
        if isinstance(left, str) or isinstance(right, str):
            if op != '+':
                raise ValueError(f"Operator '{op}' not allowed for text operands")
            return ('' if left is None else str(left)) + ('' if right is None else str(right))

        # numeric ops
        if left is None or right is None:
            raise ValueError("Operations with NULL are not allowed")

        if not (is_float_value(left) and is_float_value(right)):
            raise ValueError(
                f"Incompatible types for operator '{op}': {type(left)}, {type(right)}"
            )

        l = float(left)
        r = float(right)

        if op == '+':
            return l + r
        if op == '-':
            return l - r
        if op == '*':
            return l * r
        if op == '/':
            return l / r
        raise ValueError(f"Unsupported operator '{op}'")

    def parse_factor():
        tok = peek()
        if tok is None:
            raise ValueError("Unexpected end of expression")

        ttype, value = tok

        # [Column Name]
        if ttype == 'COL':
            advance()
            name = value[1:-1]
            if name not in context:
                raise KeyError(f"Column '{name}' not found in row")
            return context[name]

        # numbers
        if ttype == 'NUMBER':
            advance()
            return float(value) if '.' in value else int(value)

        # quoted strings
        if ttype == 'STRING':
            advance()
            return value[1:-1]

        # IDENT (DATE(...), bare names, null)
        if ttype == 'IDENT':
            name = value
            nxt = tokens[pos + 1] if pos + 1 < len(tokens) else None

            # function call, e.g. DATE(...), IF(...)
            if nxt and nxt[0] == '(':
                func_name = name.upper()
                advance()  # IDENT
                advance()  # '('
                args = []
                if peek() and peek()[0] != ')':
                    args.append(parse_expression())
                    while peek() and peek()[0] == 'COMMA':
                        advance()  # skip comma
                        args.append(parse_expression())
                if not peek() or peek()[0] != ')':
                    raise ValueError("Missing closing parenthesis in function call")
                advance()  # ')'

                if func_name == "DATE":
                    if len(args) != 1:
                        raise ValueError("DATE() expects exactly 1 argument")
                    arg = args[0]
                    if arg is None:
                        return None
                    if hasattr(arg, "date"):
                        return arg.date().isoformat()
                    if isinstance(arg, str):
                        return arg.split(" ")[0]
                    return str(arg)

                if func_name == "IF":
                    if len(args) != 3:
                        raise ValueError("IF() expects exactly 3 arguments (condition, true_value, false_value)")
                    condition, true_val, false_val = args
                    return true_val if condition else false_val

                raise ValueError(f"Unknown function '{func_name}'")

            # bare variable
            advance()
            if name.lower() == 'null':
                return None
            if name not in context:
                raise KeyError(f"Name '{name}' not found in row")
            return context[name]

        # parentheses
        if ttype == '(':
            advance()
            v = parse_expression()
            if not peek() or peek()[0] != ')':
                raise ValueError("Missing closing parenthesis")
            advance()
            return v

        # unary +/- 
        if ttype == '+':
            advance()
            return parse_factor()
        if ttype == '-':
            advance()
            v = parse_factor()
            if not is_float_value(v):
                raise ValueError("Unary '-' only allowed on numeric values")
            return -float(v)

        raise ValueError(f"Unexpected token {tok} in factor")

    def parse_term():
        v = parse_factor()
        while True:
            tok = peek()
            if tok is None or tok[0] not in ('*', '/'):
                break
            op = tok[0]
            advance()
            right = parse_factor()
            v = apply_op(op, v, right)
        return v

    def parse_additive():
        v = parse_term()
        while True:
            tok = peek()
            if tok is None or tok[0] not in ('+', '-'):
                break
            op = tok[0]
            advance()
            right = parse_term()
            v = apply_op(op, v, right)
        return v

    def apply_comp(op, left, right):
        if op == '>=': return left >= right
        if op == '<=': return left <= right
        if op == '==': return left == right
        if op == '!=': return left != right
        if op == '>':  return left > right
        if op == '<':  return left < right
        raise ValueError(f"Unknown comparison operator '{op}'")

    def parse_expression():
        v = parse_additive()
        tok = peek()
        if tok and tok[0] == 'COMP':
            comp_op = tok[1]
            advance()
            right = parse_additive()
            v = apply_comp(comp_op, v, right)
        return v

    result = parse_expression()
    if pos != len(tokens):
        raise ValueError("Unexpected extra tokens in formula")
    return result


def resolve_condition_output_generic(value_def, row: RowLike):
    """
    Shared logic for Condition.result_value / else_value:
    - If scalar -> return as-is.
    - If dict with keys:
          IS_FORMULA, DIRECT_VALUE, FORMULA
      then:
          IS_FORMULA == "Y" -> evaluate FORMULA on the row
          otherwise          -> DIRECT_VALUE
    """
    if not isinstance(value_def, dict):
        return value_def

    is_formula = str(value_def.get("IS_FORMULA", "N")).upper() == "Y"
    if is_formula:
        formula = value_def.get("FORMULA")
        if not formula:
            return value_def.get("DIRECT_VALUE")
        return evaluate_formula_expression_generic(formula, row)
    else:
        return value_def.get("DIRECT_VALUE")


def extract_sorted_categories(grouped_mappings: dict[Category, list]) -> list[Category]:
    categories: list[Category] = list(grouped_mappings.keys())
    return sorted(categories, key=lambda category: category.group_num)


def extract_sheets(ordered_categories: list[Category]) -> list[str]:
    """
    From the provided categories, returns an ordered list of the sheet names,
    based on the map group ordering of the categories.
    """
    sheets: list[str] = []
    for category in ordered_categories:
        if category.sheet not in sheets:
            sheets.append(category.sheet)
    return sheets


def trim_string_fields(df: pd.DataFrame) -> pd.DataFrame:
    df = df.map(lambda x: x.strip() if isinstance(x, str) else x)
    return df


def test_func():
    return "..."


from typing import Any
import json
from OTCFinUtils.dataverse_handler import DVHandler
from OTCFinUtils.data_structs import (
    IDParams,
    APIParams,
    Lookup,
    EntityLookups,
)

def _escape_odata_condition(condition: str) -> str:
    """
    Escapes single quotes inside OData filter condition values.
    e.g. "col eq ''20251001003504'" -> "col eq '''20251001003504'"
    Normal values without quotes are left unchanged.
    """
    EQ_PREFIX = " eq '"
    if EQ_PREFIX not in condition or not condition.endswith("'"):
        return condition

    prefix_end = condition.index(EQ_PREFIX) + len(EQ_PREFIX)
    prefix = condition[:prefix_end]
    value = condition[prefix_end:-1]  # the value without wrapping quotes
    value = value.replace("'", "''")
    return f"{prefix}{value}'"


def _extract_key_api_params(dvh: DVHandler, id_params: IDParams) -> APIParams:
    """
    Helper function for creating appropriate DV API parameters
    from the provided id parameters.
    """
    escaped_conditions = [_escape_odata_condition(c) for c in id_params.conditions]
    conditions = " and ".join(escaped_conditions)
    url = dvh._create_url(
        domain=id_params.table,
        select=id_params.id_column,
        filter=conditions,
    )
    return APIParams(url=url, method="GET")


def _extract_batch_response_objects(batch_response: str) -> list:
    """
    A general helper function. Can be used for any type of batch response.
    """
    try:
        objects = []
        responses = batch_response.split("--batchresponse")
        responses = responses[1:-1]
        for response in responses:
            start_index = response.find('{')
            end_index = response.rfind('}') + 1
            json_string = response[start_index:end_index]
            parsed_json = json.loads(json_string)
            objects.append(parsed_json["value"])
        return objects
    
    except Exception as exception:
        raise Exception(
            f"Error while extracting batch response objects. "
            f"Error: {exception}. "
            f"Response: {batch_response}."
        )


def _extract_object_ids(objects: list, id_params_buffer: list[IDParams]) -> list[str]:
    """
    Returns a list of IDs from a list of json response objects.
    Uses the id-parameters-buffer to check the form of the object.
    
    If it's a lookup-key, the id will be in a different location,
    than if it were a regular key.
    
    If the entity is new, return None.
    """
    ids = []
    for object, id_params in zip(objects, id_params_buffer):
        if len(object):
            data = object[0]
        else:
            ids.append(None)
            continue
        
        # This means we have a regular key and the entity exists in the DV
        if id_params.id_column in data:
            id = data[id_params.id_column]
        # This means we have a regular key and the entity is new
        else:
            id = None
        
        ids.append(id)
    return ids


def _extract_object_lookups(objects: list, lookup_params: list[list[IDParams]]) -> list[EntityLookups]:
    """
    Helper function for getting the lookups from the batch responses.
    """
    lookups = []
    from_index = 0
    
    for entity_lookup_params in lookup_params:
        lookup_number = len(entity_lookup_params)
        to_index = from_index + lookup_number
        entity_objects = objects[from_index : to_index]
        entity_lookups = _extract_curr_object_lookups(entity_objects, entity_lookup_params)
        lookups.append(entity_lookups)
        from_index += lookup_number

    return lookups


def _extract_curr_object_lookups(entity_objects: list, entity_params: list[IDParams]) -> list[Lookup]:
    """
    Helper function. One entity can have multiple lookups. Because of this,
    it's better to group this logic into a separate function.
    """
    if len(entity_objects) != len(entity_params):
        raise RuntimeError("The lenghts of the objects and parameters do not match.")

    entity_lookups = []
    for object, params in zip(entity_objects, entity_params):
        
        if not len(object):
            lookup = Lookup(params.bind_column, params.table, id=None)
        else:
            data = object[0]
            id = data[params.id_column]
            lookup = Lookup(params.bind_column, params.table, id)
        entity_lookups.append(lookup)
    
    return entity_lookups


def get_buffer_lookups(dvh: DVHandler, lookup_params_buffer: list[list[IDParams]]) -> list[EntityLookups]:
    """
     Uses the provided buffer to get all of the ids of the lookups.
    We need the ids of the lookups in order to properly upsert the
    data to DV — chunked into batches of max 1000 APIParams.
    
    Chunked version, but preserves empty lookup rows to match original behavior.
    """
    max_batch_size = 1000
    all_results: list[EntityLookups] = []
 
    current_batch_params: list[APIParams] = []
    current_lookup_chunk: list[list[IDParams]] = []
    current_count = 0
 
    for lookup_row in lookup_params_buffer:
        # build API params for this row
        row_api_params = [_extract_key_api_params(dvh, lp) for lp in lookup_row]
        row_count = len(row_api_params)
 
        # if exceeding limit → execute current chunk
        if current_count + row_count > max_batch_size and current_batch_params:
            batch_response = dvh._execute_batch_operation(current_batch_params)
            response_objects = _extract_batch_response_objects(batch_response)
            result = _extract_object_lookups(response_objects, current_lookup_chunk)
            all_results.extend(result)
 
            current_batch_params = []
            current_lookup_chunk = []
            current_count = 0
 
        # add this row even if empty
        current_batch_params.extend(row_api_params)
        current_lookup_chunk.append(lookup_row)
        current_count += row_count
 
    # last chunk
    if current_lookup_chunk:
        if current_batch_params:
            batch_response = dvh._execute_batch_operation(current_batch_params)
            response_objects = _extract_batch_response_objects(batch_response)
            result = _extract_object_lookups(response_objects, current_lookup_chunk)
        else:
            # No API calls but need to preserve empty structure
            result = [[] for _ in current_lookup_chunk]
        all_results.extend(result)
 
    return all_results


def get_buffer_ids(dvh: DVHandler, id_params_buffer: list[IDParams]) -> list[str]:
    """
    Uses the provided buffer to get all of the ids of the entities.
    We need the ids of the entities in order to properly upsert the
    data to DV.
    """
    batch_params: list[APIParams] = []
    
    for id_params in id_params_buffer:
        api_params = _extract_key_api_params(dvh, id_params)
        batch_params.append(api_params)
    
    batch_response = dvh._execute_batch_operation(batch_params)
    response_objects = _extract_batch_response_objects(batch_response)
    
    result = _extract_object_ids(response_objects, id_params_buffer)
    
    return result


from http import HTTPStatus as status
from enum import Enum
from dataclasses import dataclass, field
import json
from typing import Optional, Any
from collections import namedtuple


@dataclass
class ResponseIds:
    """
    Used as a return value from the upsert operation on the DV.
    Contains the table names and ids of the entities that were 
    updated and created, stored separately.
    """
    def __init__(self) -> None:
        self.updated_entity_ids: dict[str, set[str]] = dict()
        self.created_entity_ids: dict[str, set[str]] = dict()
        self.failed_operations: list[str] = list()

    
    def append(self, status_code: int, id: str, table: str) -> None:
        """
        Depending on the status code appends the id and 
        table name to the create / update category.
        """
        if status_code == status.OK.value and table in self.updated_entity_ids:
            self.updated_entity_ids[table].add(id)
        
        elif status_code == status.OK.value and table not in self.updated_entity_ids:
            self.updated_entity_ids[table] = set()
            self.updated_entity_ids[table].add(id)
        
        elif status_code == status.CREATED.value and table in self.created_entity_ids:
            self.created_entity_ids[table].add(id)
        
        elif status_code == status.CREATED.value and table not in self.created_entity_ids:
            self.created_entity_ids[table] = set()
            self.created_entity_ids[table].add(id)
        
        return
    

    def merge(self, delta_ids: "ResponseIds") -> None:
        """
        Merges the values from the provided object onto the
        the values of the calling object.
        """
        for table, ids in delta_ids.created_entity_ids.items():
            if table not in self.created_entity_ids:
                self.created_entity_ids[table] = set()
            self.created_entity_ids[table].update(ids)

        for table, ids in delta_ids.updated_entity_ids.items():
            if table not in self.updated_entity_ids:
                self.updated_entity_ids[table] = set()
            self.updated_entity_ids[table].update(ids)

        self.failed_operations.extend(delta_ids.failed_operations)

        return
    

    def json_updated_entities(self) -> dict[str, list]:
        result = dict()
        for table, ids in self.updated_entity_ids.items():
            result[table] = list(ids)
        return result
    

    def json_created_entities(self) -> dict[str, list]:
        result = dict()
        for table, ids in self.created_entity_ids.items():
            result[table] = list(ids)
        return result
    

    def num_updated_entities(self) -> int:
        total = 0
        for ids in self.updated_entity_ids.values():
            total += len(ids)
        return total
    
    def num_created_entities(self) -> int:
        total = 0
        for ids in self.created_entity_ids.values():
            total += len(ids)
        return total
    
    def num_failed_operations(self) -> int:
        return len(self.failed_operations)
    

class FileSource(Enum):
    DV = "DV"
    SP = "SP"


@dataclass
class APIParams:
    """
    Basic helper data structure for making API calls.
    """
    url: str
    method: str
    data: dict | None = None


class QueryOperator(Enum):
    AND = "and"
    OR = "or"


@dataclass
class QueryParams:
    table: str
    columns: Optional[list[str]] = None
    conditions: Optional[list[str]] = None
    operator: QueryOperator = QueryOperator.AND
    top_num: Optional[int] = None


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    DELETE = "DELETE"


@dataclass
class KeyLookupParams:
    lookup_id_column: str
    lookup_id: Optional[str]


EntityKeyLookupParams = dict[str, KeyLookupParams]


@dataclass
class IDParams:
    """
    A group of parameters used for querying id info of entities.
    """
    table: Optional[str] = None
    id_column: Optional[str] = None
    conditions: list[str] = field(default_factory=list)
    
    # If we have a lookup
    bind_column: Optional[str] = None
    

EntityLookupParams = list[IDParams]

"""
Used to divide and sort the file loading mappings
sql_table: None for Dataverse upsert, or SQL table name for SQL upsert
"""
Category = namedtuple("Category", ["group_num", "table", "sheet", "sql_table"])

"""
Used for storing the id info of the lookup entity
"""
Lookup = namedtuple("Lookup", ["bind_column", "table", "id"])

"""
Needed because each entity can have several lookups to other entities
"""
EntityLookups = list[Lookup]


class MappingFields:
    LOOKUP_TABLE = "new_lookuptable"
    LOOKUP_RELATIONSHIP = "new_lookuprelationship"
    LOOKUP_COLUMN = "new_lookupcolumn"
    DV_COLUMN = "new_dataversecolumnname"
    DV_TABLE = "new_dataversetablenametext"
    FILE_COLUMN = "new_filecolumnname"
    DEFAULT_VALUE = "new_defaultvalue"
    DATA_TYPE = "new_dataversedatatype"
    IS_LOOKUP = "new_islookup"
    IS_KEY = "new_iskey"
    IS_CHOICE = "new_ischoice"
    CHOICE_DICTIONARY = "new_choicedictionary"
    CONDITION = "new_condition"
    ALTERNATIVE_FILE_COLUMN = "new_alternativefilecolumnname"
    EXCLUDE_VALUE = "new_excludevalue"
    MAP_GROUP_ORDER = "new_mapgrouporder"
    SHEET = "new_sheetname"
    UPSERT_TO_SQL = "new_upserttosql"
    SQL_TABLE_NAME = "new_upserttosqltablename"
    SQL_COLUMN_NAME = "new_sqlcolumnname"


@dataclass
class Mapping:
    dv_column: str
    dv_table: str
    file_column: str
    map_group_order: int
    sheet: str
    default_value: Optional[str] = None
    data_type: Optional[str] = None
    lookup_table: Optional[str] = None
    lookup_relationship: Optional[str] = None
    lookup_column: Optional[str] = None
    is_lookup: bool = False
    is_key: bool = False
    is_choice: bool = False
    choice_dictionary: Optional[str] = None
    condition: Optional[str] = None
    alternative_file_column: Optional[str] = None
    exclude_value: Optional[str] = None


class Operator(Enum):
    EQUALS = "EQUALS"
    NOT_EQUALS = "NOT_EQUALS"
    IS_NULL = "IS_NULL"
    IS_NOT_NULL = "IS_NOT_NULL"
    SMALLER = "SMALLER"
    GREATER = "GREATER"
    GREATER_OR_EQUAL = "GREATER_OR_EQUAL"
    SMALLER_OR_EQUAL = "SMALLER_OR_EQUAL"


@dataclass
class Condition:
    column: str
    operator: Operator
    is_static_value: bool
    compare_value: Any
    result_value: Any
    else_value: Any

    @classmethod
    def create(cls, obj: dict) -> "Condition":
        def _parse(val):
            if not isinstance(val, str):
                return val
            try:
                return json.loads(val)
            except Exception:
                return val

        return cls(
            column=obj["column"],
            operator=Operator(obj["operator"]),
            is_static_value=obj["is_static_value"],
            compare_value=obj["compare_value"],
            result_value=_parse(obj.get("result_value")),
            else_value=_parse(obj.get("else_value")),
        )

    

DEFAULT_SHEET_CHUNK_SIZE = 1000
SHEETS_RULE_TYPE_CHOICE = 12
NOT_DEFINED = "NOT DEFINED"

_TOKEN_SPEC = [
    ('COL',     r'\[[^\]]+\]'),
    ('NUMBER',  r'\d+(\.\d+)?'),
    ('IDENT',   r'[A-Za-z_][A-Za-z0-9_]*'),
    ('STRING',  r"'[^']*'|\"[^\"]*\""),
    ('COMP',    r'>=|<=|==|!=|>|<'),      
    ('COMMA',   r','),
    ('OP',      r'[+\-*/()]'),
    ('SKIP',    r'\s+'),
    ('MISMATCH',r'.'),
]

_TOKEN_REGEX = '|'.join(f'(?P<{name}>{regex})' for name, regex in _TOKEN_SPEC)




from OTCFinUtils.dataverse_handler import (
    DVHandler,
    get_dv_handler,
    set_up_dv_handler,
)
from OTCFinUtils.buffer import Buffer
from OTCFinUtils.data_structs import (
    ResponseIds,
    APIParams,
    HTTPMethod,
    KeyLookupParams,
    EntityKeyLookupParams,
    IDParams,
    EntityLookupParams,
    Category,
    Lookup,
    EntityLookups,
    MappingFields,
    Operator,
    Condition,
    DEFAULT_SHEET_CHUNK_SIZE,
    SHEETS_RULE_TYPE_CHOICE,
    FileSource,
)
from OTCFinUtils.mapper_loading import (
    _extract_key_api_params,
    _extract_batch_response_objects,
    _extract_object_ids,
    _extract_object_lookups,
    _extract_curr_object_lookups,
    get_buffer_lookups,
    get_buffer_ids,
)
from OTCFinUtils.mapper import DataMapper
from OTCFinUtils.security import (
    get_client_credential,
    get_dataverse_token,
    get_graph_token,
)
from OTCFinUtils.sharepoint import (
    load_document,
    get_connection_details,
    create_excel_document,
    update_excel,
)
from OTCFinUtils.utils import (
    extract_segment_object_line,
    extract_segment_response_id,
    extract_segment_status_code,
    extract_segment_table_name,
    process_data_type,
    extract_lookup_column,
    extract_file_data,
    data_equals_exclude_value,
    trim_string_values,
    is_date,
    extract_date_string,
    choice_value,
    normalize_choice_map,
    default_value,
    cast_data,
    get_skip_flag,
    data_is_empty,
    evaluate_condition,
    is_valid,
    group_mappings,
    extract_sorted_categories,
    extract_sheets,
)


from http import HTTPStatus as status
from typing import Optional
from dotenv import load_dotenv
from os import getenv
import json
import urllib.parse
import requests
from OTCFinUtils.security import get_dataverse_token, is_jwt_expired
from OTCFinUtils.data_structs import (
    APIParams, 
    HTTPMethod,
    QueryParams, 
    ResponseIds,
)
from OTCFinUtils.utils import (
    extract_segment_response_id, 
    extract_segment_status_code, 
    extract_segment_table_name
)
from functools import wraps
from OTCFinUtils.utils import chunked_list
import copy
import time
import logging

class DVHandler:
    _BATCH_BOUNDARY: str = "batch_boundary"
    _BATCH_CHUNK_SIZE: int = 900
    _NEXT_PAGE_KEY: str = "@odata.nextLink"
    _instance = None

    def _get(self, url, max_retries=3, initial_delay=5, **kwargs):
        """GET request with retry on transient network/DNS errors."""
        delay = initial_delay
        for attempt in range(max_retries + 1):
            try:
                return self._get(url, **kwargs)
            except Exception as e:
                error_str = str(e).lower()
                is_network_error = any(x in error_str for x in [
                    'name resolution', 'temporary failure', 'connection',
                    'timeout', 'errno -3', 'errno 11001', 'getaddrinfo',
                    'network unreachable'
                ])
                if is_network_error and attempt < max_retries:
                    logging.warning(
                        f"Dataverse GET failed (attempt {attempt + 1}/{max_retries + 1}): {e}. "
                        f"Retrying in {delay}s..."
                    )
                    time.sleep(delay)
                    delay = min(delay * 2, 60)
                else:
                    raise


    def __new__(cls, dataverse_url: Optional[str] = None, user_id: Optional[str] = None) -> "DVHandler":
        if dataverse_url is None:
            load_dotenv()
            dataverse_url = getenv("DATAVERSE_DEFAULT_URL", "")

        if not cls._instance or f"{urllib.parse.urlparse(getattr(cls._instance, 'api_url', '')).scheme}://{urllib.parse.urlparse(getattr(cls._instance, 'api_url', '')).netloc}".rstrip('/') != dataverse_url.rstrip('/'):
            cls._instance = super(DVHandler, cls).__new__(cls)
            cls._instance.__init__(dataverse_url, user_id)
        else:
            # For existing instance with the same base URL, make sure token is still valid
            try:
                token_value = getattr(cls._instance, "token", None)
                if token_value is None or is_jwt_expired(token_value):
                    cls._instance.__init__(dataverse_url, user_id, force_refresh=True)
            except Exception:
                cls._instance.__init__(dataverse_url, user_id, force_refresh=True)
        
        return cls._instance # type: ignore
    

    def __init__(self, dataverse_url: str, user_id: Optional[str] = None, force_refresh: bool = False) -> None:
        if getattr(self, "_initialized", False) and not force_refresh:
            return

        self.api_url = f"{dataverse_url}/api/data/v9.2"
        self.token = get_dataverse_token(dataverse_url)
        self.headers = {'Authorization': f'Bearer {self.token}', 'Content-Type': 'application/json'}
        self.user_id = user_id
        self._initialized = True
 
    def get_entities_list(
        self,
        table: str,
        filter: Optional[str] = None,
        select: Optional[str] = None,
        expand: Optional[str] = None,
        top: Optional[int] = None,
        parse_filter: bool = True,
        order_by: Optional[str] = None,
        max_attempts: int = 5,
    ) -> list[dict]:
        api_url = self._create_url(
            domain=table,
            select=select,
            filter=filter,
            expand=expand,
            top=top,
            parse_filter=parse_filter,
            order_by=order_by,
        )
        response = self._get(api_url,headers=self.headers)
        
        if not response.status_code == status.OK.value:
            raise RuntimeError(f"Could not query entities for: {table}. Error: {response.text}")
        
        response_data = response.json()
        entities: list[dict] = response_data["value"]

        if DVHandler._NEXT_PAGE_KEY in response_data:
            next_page_link = response_data[DVHandler._NEXT_PAGE_KEY]
            next_page_entities = self.get_next_page(next_page_link)
            entities.extend(next_page_entities)

        return entities

        
    @wraps(get_entities_list)
    def chunked_get_entities_list(self, *args, **kwargs) -> list[dict]:
        """
        args: same as get_entities_list, but filter only contains stable filter
            (non-changing conditions that we want to apply for every chunk call)

        extra keyword arguments:
            - stable_to_chunked_connector: 'or' | 'and'
            - chunk_filter: list[str]   # list of condition strings
            - chunk_filter_join_with: str  # 'or' | 'and'
            - chunk_size: int
        """

        def _normalize_op(op: str) -> str:
            op = (op or "").strip().lower()
            if op not in ("and", "or"):
                raise ValueError(f"Invalid logical operator: {op!r}. Use 'and' or 'or'.")
            return op

        def _join_filters(parts, op: str) -> str:
            op = _normalize_op(op)
            parts = [p.strip() for p in parts if p and str(p).strip()]
            if not parts:
                return ""
            if len(parts) == 1:
                return parts[0]
            return f"({f' {op} '.join(parts)})"

        def _remove_unbalanced_parentheses(items: list[str]) -> list[str]:
            """
            Remove unbalanced parentheses from the beginning and end of the list.
            """
            if not items:
                return items
            first = items[0]
            if first.startswith("(") and ")" not in first:
                first = first[1:]  # remove leading '('
            
            last = items[-1]
            if last.endswith(")") and "(" not in last:
                last = last[:-1]  # remove trailing ')'
            return [first, *items[1:-1], last]

        # Extract required params (don’t mutate caller’s kwargs beyond pops here)
        try:
            stable_to_chunked_connector = _normalize_op(kwargs.pop("stable_to_chunked_connector"))
            chunk_filter = kwargs.pop("chunk_filter")  # now a list[str]
            chunk_filter = _remove_unbalanced_parentheses(chunk_filter)
            chunk_filter_join_with = _normalize_op(kwargs.pop("chunk_filter_join_with"))
            chunk_size = int(kwargs.pop("chunk_size"))
        except KeyError as e:
            raise TypeError(f"Missing required param: {e.args[0]}")
        except ValueError as e:
            # Covers invalid op or non-int chunk_size
            raise

        if chunk_size <= 0:
            raise ValueError("chunk_size must be > 0")
        if not isinstance(chunk_filter, list) or not all(isinstance(c, str) for c in chunk_filter):
            raise TypeError("chunk_filter must be a list of strings")

        stable_filter = (kwargs.get("filter") or "").strip()

        # Normalize and prune empty entries
        chunk_filter_list = [c.strip() for c in chunk_filter if isinstance(c, str) and c.strip()]
        if not chunk_filter_list:
            # Nothing to chunk; just call once with stable filter

            print(f"\033[31m calling get_entities_list with args: {args} and kwargs:{call_kwargs} \033[0m")
            print("--------------------------")
            return self.get_entities_list(*args, **kwargs)

        entities: list[dict] = []

        for chunk in chunked_list(chunk_filter_list, chunk_size):
            chunk_part = _join_filters(chunk, chunk_filter_join_with)
            combined = _join_filters([stable_filter, chunk_part], stable_to_chunked_connector) if stable_filter else chunk_part

            call_kwargs = copy.deepcopy(kwargs)
            call_kwargs["filter"] = combined
            print(f"\033[31m calling get_entities_list with args: {args} and kwargs:{call_kwargs} \033[0m")
            print("--------------------------")
            results = self.get_entities_list(*args, **call_kwargs)
            entities.extend(results)

        return entities

    def _create_url(
        self, 
        domain: Optional[str] = None, 
        select: str|list|None = None, 
        filter: Optional[str] = None, 
        expand: Optional[str] = None, 
        top: Optional[int] = None,
        parse_filter: bool = True,
        order_by: Optional[str] = None,
    ) -> str:
        """
        A general helper function. Can be used extensively.
        """
        url = f"{self.api_url}/{domain}?"
        params: list[str] = []
        
        if isinstance(select, str):
            params.append(f"$select={select}")
        
        elif isinstance(select, list):
            columns = ",".join(select)
            params.append(f"$select={columns}")

        if filter and parse_filter:
            params.append(f"$filter={urllib.parse.quote(filter)}") if filter else None
        elif filter and not parse_filter:
            params.append(f"$filter={filter}") if filter else None
        
        params.append(f"$expand={expand}") if expand else None
        params.append(f"$top={top}") if top else None
        params.append(f"$orderby={order_by}") if order_by else None
        url += "&".join(params)
        return url
    

    def get_next_page(self, page_link: str) -> list[dict]:
        response = self._get(page_link, headers=self.headers)

        if response.status_code != status.OK.value:
            raise RuntimeError(f"Cannot get page data for link: {page_link}.")

        page_data = response.json()
        entities: list[dict] = page_data["value"]
        
        if DVHandler._NEXT_PAGE_KEY in page_data:
            next_page_link = page_data[DVHandler._NEXT_PAGE_KEY]
            next_page_entities = self.get_next_page(next_page_link)
            entities.extend(next_page_entities)
        
        return entities
    

    @staticmethod
    def create_expand_query_string(lookup_name: str, select: Optional[str]=None, filter: Optional[str]=None) -> str:
        params = []
        
        if select:
            select = f"$select={select}"
            params.append(select)
        if filter:
            filter = f"$filter={filter}"
            params.append(filter)
        
        params = ";".join(params)

        return f"{lookup_name}({params})"
    

    def get_table_definition(self, table: str, fields: Optional[str] = None) -> dict:
        """
        Returns the table definition information of the given table.
        """
        url = self._create_url(f"EntityDefinitions(LogicalName='{table}')", fields)
        response = self._get(url=url, headers=self.headers)

        if response.status_code != status.OK.value:
            raise RuntimeError(f"Could not get table definition: {response.text}")

        return response.json()
    

    # TODO - review and test this function
    
    def get_columns_names(self, table_name: str) -> list[str]:
        expand = "Attributes($select=DisplayName,LogicalName,AttributeType,SchemaName),"
        expand += "Keys($select=KeyAttributes)"
        
        url = self._create_url(
            domain=f"EntityDefinitions(LogicalName='{table_name}')",
            select="LogicalName",
            filter=None,
            expand=expand,
        )
 
        response = self._get(url, headers=self.headers)
        keys = []
        
        if response.status_code != status.OK.value:
            raise RuntimeError(f"Could not get column names for {table_name}. Response: {response.text}")
        
        data = response.json()
        keysData = data.get("Keys", [])
        for item in keysData:
            if "KeyAttributes" in item:
                keys.extend(item["KeyAttributes"])  

        attributes = data.get("Attributes", [])
        filtered_attributes = []
        
        for attribute in attributes:
            if not attribute["LogicalName"].startswith("new_") and not attribute["LogicalName"] == "statecode":
                continue

            if "DisplayName" not in attribute or not attribute["DisplayName"]:
                continue

            display_name_field = attribute["DisplayName"]
            
            if (
                "UserLocalizedLabel" not in display_name_field or 
                not display_name_field["UserLocalizedLabel"] or 
                not display_name_field["UserLocalizedLabel"]["Label"]
            ):
                continue

            is_key = attribute["LogicalName"] in keys
            is_lookup = attribute["AttributeType"] == "Lookup"
            is_choice = attribute["AttributeType"] == "Picklist"
            is_bool = attribute["AttributeType"] == "Boolean"
            
            lookup_table_name = None
            lookup_column_name = None
            lookup_relationship = None
            
            choice_name = None
            choice_default_value= None
            choice_dictionary= None
            
            if is_lookup:
                lookup_details = self.get_lookup_details(
                    table_name, attribute["LogicalName"]
                )
                lookup_table_name = lookup_details["lookup_table_name"]
                lookup_column_name = lookup_details["lookup_column_name"]
                lookup_relationship = lookup_details["lookup_relationship"]
            
            elif is_choice:
                choice_details = self.get_choice_details(table_name, attribute["LogicalName"])
                choice_name = choice_details["choice_name"]
                choice_default_value= choice_details["choice_default_value"]
                choice_dictionary= choice_details["choice_dictionary"]
            
            elif is_bool:  
                bool_dct= []
                bool_dct.append({"Value": True, "DV_Display": "Yes", "Excel_Value": "Yes"})
                bool_dct.append({"Value": False, "DV_Display": "No", "Excel_Value": "No"})
                choice_dictionary= json.dumps(bool_dct)
            
        
            filtered_attributes.append({
                "LogicalName": attribute["SchemaName"] if is_lookup else attribute["LogicalName"],
                "DisplayName": attribute["DisplayName"]["UserLocalizedLabel"]["Label"],
                "AttributeType": attribute["AttributeType"],
                "IsKey": is_key,
                "IsLookup": is_lookup,
                "LookupRelationship":lookup_relationship,
                "LookupTable": lookup_table_name,
                "LookupColumn": lookup_column_name,
                "IsChoice": is_choice,
                "ChoiceName": choice_name,
                "ChoiceDefaultValue": choice_default_value,
                "ChoiceDictionary": choice_dictionary,
            })

        return filtered_attributes
        

    # TODO - review and test this function
    
    def get_lookup_details(self, table_name: str, logical_name: str) -> dict:
        url = self._create_url(
            domain=f"EntityDefinitions(LogicalName='{table_name}')/ManyToOneRelationships",
            select="ReferencedEntityNavigationPropertyName,ReferencedEntity,ReferencedAttribute,ReferencingAttribute",
            filter=f"ReferencingAttribute eq '{logical_name}'"
        )

        response = self._get(url, headers=self.headers)

        if response.status_code != status.OK.value:
            raise RuntimeError(f"Could not get lookup details for {table_name}. Response: {response.text}")
        
        data = response.json()
        if not data["value"]:
            raise RuntimeError(f"No data received for {table_name}")
        
        relationship = data["value"][0]
        lookup_table_name = relationship["ReferencedEntity"]
        lookup_relationship = relationship["ReferencedEntityNavigationPropertyName"]

        key_url = self._create_url(
            domain=f"EntityDefinitions(LogicalName='{lookup_table_name}')/Keys",
            select="KeyAttributes",
        )
        key_response = self._get(key_url, headers=self.headers)
        lookup_column_name = None
        
        if key_response.status_code != status.OK.value:
            raise RuntimeError(f"Could not get lookup details for {table_name}. Response: {key_response.text}")
        
        key_data = key_response.json()
        if not key_data["value"]:
            raise RuntimeError(f"Could not get lookup details for {table_name}. No 'value' field in response.")

        keys_data = key_data["value"][0]                    
        lookup_column_name = str(keys_data["KeyAttributes"])

        return {
            "lookup_table_name": lookup_table_name,
            "lookup_column_name": lookup_column_name,
            "lookup_relationship": lookup_relationship,
        }
        

    # TODO - review and test this function
    
    def get_choice_details(self, table_name: str, logical_name: str) -> dict:
        url = f"{self.api_url}/EntityDefinitions(LogicalName='{table_name}')"
        url += f"/Attributes(LogicalName='{logical_name}')"
        url += "/Microsoft.Dynamics.CRM.PicklistAttributeMetadata?"
        url += "$select=LogicalName,DefaultFormValue&"
        url += "$expand=GlobalOptionSet($select=Name,Options)"
        
        response = self._get(url, headers=self.headers)
        choice_default_value =  None
        
        if response.status_code != status.OK.value:
            raise RuntimeError(f"Could not get choice details for {table_name}. Response: {response.text}")
        
        data = response.json()
        if data["DefaultFormValue"]:
            choice_default_value =  None if data["DefaultFormValue"]  == -1 else data["DefaultFormValue"] 
        
        if not data["GlobalOptionSet"]:
            raise RuntimeError(f"Could not get choice details for {table_name}. Global Option Set = {data['GlobalOptionSet']}")
        
        choice = data["GlobalOptionSet"]
        choiceName = choice["Name"]
        choiceOptions = choice["Options"]

        result = []
        
        for option in choiceOptions:
            result.append({
                "Value": option["Value"], 
                "DV_Display": option["Label"]["UserLocalizedLabel"]["Label"],
                "Excel_Value": option["Label"]["UserLocalizedLabel"]["Label"]}
            ) 
        
        json_result = json.dumps(result) 
        
        return {
            "choice_name": choiceName,
            "choice_default_value": choice_default_value,
            "choice_dictionary": json_result
        }
    

    def get_choice_mappings(self, logical_choice_name: str) -> dict:
        url = self._create_url(
            domain="stringmaps",
            select=["value", "attributevalue"],
            filter=f"attributename eq '{logical_choice_name}'",
            parse_filter=False
        )
        result = self._get(url,headers=self.headers).json()["value"]
        mappings = {}
        
        for item in result:
            key = item["value"]
            value = item["attributevalue"]
            if key not in mappings:
                mappings[key] = value
        
        return mappings


    def upsert_bulk(self, data: list[dict], no_response=False) -> ResponseIds:
        """
        Executes a bulk upsert operation.
        If the data has an id, then update the entity.
        If the data has no id, then insert the entity.

        If there are more than items than the chunk size, 
        splits the operation into multiple batches.
        """
        response_ids = ResponseIds()
        
        for i in range(0, len(data), DVHandler._BATCH_CHUNK_SIZE):
            from_index = i
            to_index = min(len(data),i + DVHandler._BATCH_CHUNK_SIZE)
            subset_data = data[from_index : to_index]
            subset_response_ids = self._upsert_bulk_subset(subset_data, no_response=no_response)
            response_ids.merge(subset_response_ids)
            
        return response_ids

    
    def _upsert_bulk_subset(self, data: list[dict], no_response=False) -> ResponseIds:
        """
        Used as a helper function for the main bulk-upsert function.
        """
        batch_params: list[APIParams] = []
        id_columns: list[str] = []
        
        for record in data:
            entity_id = record.pop("entity_id")
            table_name: str = record.pop("table_name")
            id_column: str = record.pop("id_column", "")
            id_columns.append(id_column)
            
            if entity_id:
                item_url = self._create_url(domain=f"{table_name}({entity_id})", select=id_column)
                method = HTTPMethod.PATCH.value
            else:
                item_url = self._create_url(domain=table_name, select=id_column)
                method = HTTPMethod.POST.value
            
            record_params = APIParams(url=item_url, method=method, data=record)
            batch_params.append(record_params)

        batch_response = self._execute_batch_operation(batch_params, self.user_id)
        if no_response:
            return ResponseIds()
        
        response_ids = DVHandler._extract_response_ids(batch_response, id_columns)
        
        return response_ids
    
    
    @staticmethod
    def _extract_response_ids(batch_response: str, id_columns: list[str]) -> ResponseIds:
        """
        From the batch response string, returns the ids of the
        created / updated entities in a ResponseIds object, only
        if the id column names are provided.
        """
        response_ids = ResponseIds()
        response_parts = batch_response.split("--batchresponse")
        response_parts = response_parts[1:-1]
        
        for response_part, id_column in zip(response_parts, id_columns):
            response_lines = response_part.split("\n")
            object_id: Optional[str] = extract_segment_response_id(response_lines, id_column)
            status_code: int = extract_segment_status_code(response_lines)
            table: str = extract_segment_table_name(response_lines)
            
            if object_id:
                response_ids.append(status_code, object_id, table)

        return response_ids
    
    
    def _execute_batch_operation(self, batch_params: list[APIParams], user_id: Optional[str] = None) -> str:
            """
            A general helper function. Can be used for any type of batch operation.
            """
            # Ensure token is initialized
            try:
                if not hasattr(self, "token") or self.token is None:
                    self.token = get_dataverse_token(self.dataverse_url)
    
                headers = {
                    'Authorization': f'Bearer {self.token}',
                    'Content-Type': f'multipart/mixed;boundary={DVHandler._BATCH_BOUNDARY}',
                    "Prefer": 'odata.continue-on-error',
                }
                if user_id is not None:
                    headers["CallerObjectId"] = user_id
                batch_strings = []
                for curr_params in batch_params:
                    try:
                        batch_string = self._create_batch_item_string(
                            DVHandler._BATCH_BOUNDARY,
                            curr_params.method,
                            curr_params.url,
                            curr_params.data,
                            user_id,
                        )
                        batch_strings.append(batch_string)
                    except Exception as e:
                        raise RuntimeError(f"Failed to create batch item: {str(e)}")
            
                batch_strings.append(f"--{DVHandler._BATCH_BOUNDARY}--")
                batch_body = "\n".join(batch_strings) + "\r\n"
    
                api_url = f"{self.api_url}/$batch"
                response = requests.post(api_url, data=batch_body, headers=headers)
    
                if response.status_code == status.OK.value:
                    return response.text
                else:
                    raise RuntimeError(f"Batch operation was not successful: {response.text}")
            except requests.exceptions.RequestException as re:
                raise RuntimeError(f"HTTP request failed during batch operation: {str(re)}") from re
            except Exception as exc:
                raise RuntimeError(f"Unexpected error during batch operation: {str(exc)}") from exc
    
    def _create_batch_item_string(self, boundary: str, method: str, url: str, data: dict | None = None, user_id: Optional[str] = None) -> str:
        """
        A helper function for creating the batch request body.
        """
        batch_item_string = (
            f"--{boundary}\n"                      
            "Content-Type: application/http\n"     
            "Content-Transfer-Encoding: binary\n\n"
            f"{method} {url} HTTP/1.1\n"           
        )

        headers = [
            f"Authorization: Bearer {self.token}",
            "Content-Type: application/json",
            'Prefer: return=representation',
        ]
        
        if user_id is not None:
            headers.append(f"CallerObjectId: {user_id}")
        
        headers = "\n".join(headers)
        batch_item_string += f"{headers}\n\n"

        if data:
            batch_item_string += json.dumps(data)
        
        return batch_item_string

    
    def get_by_id(self, table_name: str, dataverse_id: str) -> dict:
        dv_url = f"{self.api_url}/{table_name}({dataverse_id})"
        response = self._get(dv_url,headers=self.headers)
        if response.status_code != status.OK.value:
            raise RuntimeError(f"Could not get entity by id: {dataverse_id}. Response: {response.text}")
        return response.json()
        

    def delete_entities(self, data: list[dict]) -> str:
        """
        A generic function for bulk-deleting entities 
        based on the provided list of dictionaries.
        """
        batch_params: list[APIParams] = []
        
        for record in data:
            table: str = record.pop("table")
            record_id: str = record.pop("id")
            
            url = self._create_url(f"{table}({record_id})")
            record_params = APIParams(url, "DELETE")
            
            batch_params.append(record_params)

        batch_response = self._execute_batch_operation(batch_params, self.user_id)
        
        return batch_response
    

    def update_entity(self, table: str, entity_id: str, new_data: dict, return_entity: bool = False, columns: list[str] = []) -> Optional[dict]:
        entity_location = f"{table}({entity_id})"
        headers = self.headers.copy()
        
        # For preventing create-operations
        headers["If-Match"] = "*"

        if return_entity:
            # For returning the updated entity
            headers["Prefer"] = "return=representation"
            columns_string = ",".join(columns)
            url = self._create_url(entity_location, columns_string)
            
        else:
            url = self._create_url(entity_location)

        response = requests.patch(url, headers=headers, json=new_data)

        if response.status_code not in (status.NO_CONTENT.value, status.OK.value):
            raise RuntimeError(f"Could not update entity. Error: {response.text}")
        
        if return_entity:
            return response.json()
        

    def read_bulk(self, query_params: list[QueryParams]) -> list[dict]:
        api_params: list[APIParams] = []
        for params in query_params:
            if isinstance(params.conditions, list):
                conditions = f" {params.operator.value} ".join(params.conditions)
            else:
                conditions = params.conditions

            url = self._create_url(
                params.table, 
                params.columns, 
                conditions, 
                None, 
                params.top_num
            )
            
            api_params.append(APIParams(
                url=url,
                method=HTTPMethod.GET.value,
            ))
        
        response = self._execute_batch_operation(api_params)
        entities = DVHandler._extract_batch_response_entities(response)
        return entities
    

    @staticmethod
    def _extract_batch_response_entities(batch_reponse: str) -> list[dict]:
        ENTITY_TAG = '{"@odata.context":'
        SEPARATOR = "--batchresponse"
        ERROR_TAG = '{"error":'
        STATUS_CODE_TAG = "HTTP"
        try:
            entities: list[dict] = list()
            parts = batch_reponse.split(SEPARATOR)
            parts = parts[1:-1]
            for part in parts:
                lines = part.split("\n")
                for line in lines:
                    
                    if line.startswith(STATUS_CODE_TAG):
                        status_code = DVHandler._extract_line_status_code(line)
                        if status_code != status.OK.value:
                            raise Exception(f"Error in one or more requests. Batch response:\n\n{batch_reponse}")
                    
                    elif line.startswith(ENTITY_TAG):
                        request_entities = DVHandler._append_request_entities(line)
                        entities.extend(request_entities)
                    
                    elif line.startswith(ERROR_TAG):
                        error_object = json.loads(line)
                        entities.append(error_object)
            
            return entities
        
        except Exception as exception:
            raise Exception(f"Error while extracting batch response entities: {exception}")
        
    
    @staticmethod
    def _extract_line_status_code(line: str) -> int:
        parts = line.split()
        status_code = int(parts[1])
        return status_code


    @staticmethod
    def _append_request_entities(line: str):
        result_entities: list[dict] = list()
        line_data = json.loads(line)
        request_entities = line_data["value"]
        if len(request_entities) == 0:
            result_entities.append(dict())
        else:
            result_entities.extend(request_entities)
        return result_entities
    

    @staticmethod
    def create_in_filter(column: str, values: list) -> str:
        """
        Helper function for creating a filter condition.
        """
        return (
            "Microsoft.Dynamics.CRM.In("
            "PropertyName=@p1,PropertyValues=@p2)&"
            f"@p1='{column}'&@p2={values}"
        )
    
    @classmethod
    def reset_instance(cls):
        cls._instance = None
        

def get_dv_handler() -> DVHandler:
    """
    Returns an object if it has been already created before.
    """
    return DVHandler.__new__(DVHandler)


def set_up_dv_handler(dataverse_url: str, user_id: Optional[str] = None) -> DVHandler:
    """
    Uses the Singleton pattern - creates a new object the first time.
    Then returns the same object on the second, third... call.
    """
    return DVHandler.__new__(DVHandler, dataverse_url, user_id)




from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from dotenv import load_dotenv
import base64
import json
import time
import msal
from os import getenv


def get_client_credential(secret_name: str) -> str:
    """
    Retruns the credential needed to get the token for access to the Dataverse.
    """
    load_dotenv()
    VAULT_URL = getenv("VAULT_URL", "")

    credential = DefaultAzureCredential()
    client = SecretClient(vault_url=VAULT_URL, credential=credential)
    secret_value = client.get_secret(secret_name).value
    
    if secret_value is None:
        raise RuntimeError("Could not get client credential.")
    
    return secret_value


def get_dataverse_token(dataverse_url: str) -> str:
    """
    Returns the token needed for working with the Dataverse API
    """
    scopes = [f"{dataverse_url}/.default"]
    return get_token_for_scope(scopes)
    

def get_flow_token() -> str:
    """
    Return the token needed for accessing power automate flows.
    """
    scopes = ["https://service.flow.microsoft.com//.default"]
    return get_token_for_scope(scopes)


def get_token_for_scope(scopes: list[str]) -> str:
    load_dotenv()
    TENANT_ID = getenv("TENANT_ID")
    CLIENT_ID = getenv("CLIENT_ID")
    TOKEN_AUTHORITY = getenv("DATAVERSE_TOKEN_AUTHORITY")
    DATAVERSE_SECRET_NAME = getenv("DATAVERSE_SECRET_NAME", "")
    TOKEN_CLIENT_CREDENTIAL = get_client_credential(DATAVERSE_SECRET_NAME)

    app = msal.ConfidentialClientApplication(
        client_id=CLIENT_ID,
        client_credential=TOKEN_CLIENT_CREDENTIAL,
        authority=f"{TOKEN_AUTHORITY}{TENANT_ID}",
    )
    token = app.acquire_token_for_client(scopes=scopes)
    
    if not isinstance(token, dict) or "access_token" not in token:
        raise RuntimeError(f"Could not get access token for: {scopes}")
    
    return token["access_token"]


# TODO - Add parameter to define *which* SharePoint

def get_graph_token():
    CLIENT_ID = getenv("CLIENT_ID", "")
    TENANT_ID = getenv("TENANT_ID", "")
    GRAPH_MSAL_AUTHORITY = getenv("GRAPH_MSAL_AUTHORITY", "")
    GRAPH_MSAL_SCOPE = getenv("GRAPH_MSAL_SCOPE", "")
    GRAPH_SECRET_NAME = getenv("GRAPH_SECRET_NAME", "")
    GRAPH_CLIENT_SECRET = get_client_credential(GRAPH_SECRET_NAME)

    msal_app = msal.ConfidentialClientApplication(
        client_id = CLIENT_ID,
        client_credential = GRAPH_CLIENT_SECRET,
        authority = GRAPH_MSAL_AUTHORITY + TENANT_ID,
    )
    result = msal_app.acquire_token_silent(
        scopes = [GRAPH_MSAL_SCOPE],
        account = None,
    )

    if not result:
        result = msal_app.acquire_token_for_client(scopes=[GRAPH_MSAL_SCOPE])

    if isinstance(result, dict) and "error_description" in result:
        raise Exception(result["error_description"])

    if not isinstance(result, dict) or "access_token" not in result:
        raise Exception("No Graph Access Token found")
    
    return result["access_token"]

    
def create_flow_headers() -> dict[str, str]:
    flow_token = get_flow_token()
    return {
        "Authorization": f"Bearer {flow_token}", 
        "Content-Type": "application/json",
    }


def is_jwt_expired(token: str, skew_seconds: int = 60) -> bool:
    """
    Returns True if the JWT access token is expired (or invalid), allowing a small
    clock skew window. This does not validate the signature; it only inspects
    the 'exp' claim from the payload.
    """
    try:
        parts = token.split(".")
        if len(parts) < 2:
            return True

        payload_b64 = parts[1]
        # Pad base64 string if necessary
        padding_needed = (4 - len(payload_b64) % 4) % 4
        payload_b64 += "=" * padding_needed

        payload_bytes = base64.urlsafe_b64decode(payload_b64)
        payload = json.loads(payload_bytes.decode("utf-8"))
        exp_timestamp = int(payload.get("exp", 0))
        now = int(time.time())
        return now >= (exp_timestamp - skew_seconds)
    except Exception:
        return True


from dotenv import load_dotenv
import requests
import os

class SharePointClient:
    """
    Not used anymore...

    Was used in the process of creating the initial 
    code for setting up a SharePoint connection.
    Can be removed. Did not manage to remove so far.
    """

    def __init__(self, tenant_id, client_id, client_secret, resource_url):
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.resource_url = resource_url
        self.base_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        self.headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        self.access_token = self.get_access_token()
        self.token_headers = {'Authorization': f'Bearer {self.access_token}'}
        self.graph_url = "https://graph.microsoft.com/v1.0/sites"


    def get_access_token(self):
        # Body for the access token request
        body = {
            'grant_type': 'client_credentials',
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'scope': self.resource_url + '.default'
        }
        response = requests.post(self.base_url, headers=self.headers, data=body)
        return response.json().get('access_token')  # Extract access token from the response
    

    def get_site_id(self, site_url):
        # Build URL to request site ID
        full_url = f'{self.graph_url}/{site_url}'
        response = requests.get(full_url, headers=self.token_headers)
        return response.json().get('id')  # Return the site ID


    def get_drive_id(self, site_id):
        # Retrieve drive IDs and names associated with a site
        drives_url = f'{self.graph_url}/{site_id}/drives'
        response = requests.get(drives_url, headers=self.token_headers)
        drives = response.json().get('value', [])
        return [(drive['id'], drive['name']) for drive in drives]


    def get_folder_content(self, site_id, drive_id, folder_path='root'):
        # Get the contents of a folder
        folder_url = f'{self.graph_url}/{site_id}/drives/{drive_id}/root/children'
        response = requests.get(folder_url, headers=self.token_headers)
        items_data = response.json()
        rootdir = []
        if 'value' in items_data:
            for item in items_data['value']:
                rootdir.append((item['id'], item['name']))
        return rootdir
    

    # Recursive function to browse folders
    def list_folder_contents(self, site_id, drive_id, folder_id, level=0):
        # Get the contents of a specific folder
        folder_contents_url = f'{self.graph_url}/{site_id}/drives/{drive_id}/items/{folder_id}/children'
        contents_headers = self.token_headers
        contents_response = requests.get(folder_contents_url, headers=contents_headers)
        folder_contents = contents_response.json()

        items_list = []  # List to store information

        if 'value' in folder_contents:
            for item in folder_contents['value']:
                if 'folder' in item:
                    # Add folder to list
                    items_list.append({'name': item['name'], 'type': 'Folder', 'mimeType': None})
                    # Recursive call for subfolders
                    items_list.extend(self.list_folder_contents(site_id, drive_id, item['id'], level + 1))
                elif 'file' in item:
                    # Add file to the list with its mimeType
                    items_list.append({'name': item['name'], 'type': 'File', 'mimeType': item['file']['mimeType']})

        return items_list
    

    def download_file(self, download_url, local_path, file_name):
        headers = self.token_headers
        response = requests.get(download_url, headers=headers)
        if response.status_code == 200:
            full_path = os.path.join(local_path, file_name)
            with open(full_path, 'wb') as file:
                file.write(response.content)
            print(f"File downloaded: {full_path}")
        else:
            print(f"Failed to download {file_name}: {response.status_code} - {response.reason}")
    

    def download_folder_contents(self, site_id, drive_id, folder_id, local_folder_path, level=0):
        # Recursively download all contents from a folder
        folder_contents_url = f'{self.graph_url}/{site_id}/drives/{drive_id}/items/{folder_id}/children'
        contents_headers = self.token_headers
        contents_response = requests.get(folder_contents_url, headers=contents_headers)
        folder_contents = contents_response.json()

        if 'value' in folder_contents:
            for item in folder_contents['value']:
                if 'folder' in item:
                    new_path = os.path.join(local_folder_path, item['name'])
                    if not os.path.exists(new_path):
                        os.makedirs(new_path)
                    self.download_folder_contents(site_id, drive_id, item['id'], new_path, level + 1)  # Recursive call for subfolders
                elif 'file' in item:
                    file_name = item['name']
                    file_download_url = f"{self.resource_url}/v1.0/sites/{site_id}/drives/{drive_id}/items/{item['id']}/content"
                    self.download_file(file_download_url, local_folder_path, file_name)


def test_sharepoint_client():
    # Usage of the class
    load_dotenv()
    tenant_id = os.getenv('CLIENT_ID')
    client_id = os.getenv('TENANT_ID')
    client_secret = os.getenv('GRAPH_CLIENT_SECRET')
    resource = 'https://graph.microsoft.com/'

    client = SharePointClient(tenant_id, client_id, client_secret, resource)

    site_id = "otcfin1.sharepoint.com,e6dadbba-d556-4282-bbbd-2524af8466ca,ecc3a3bf-ec4b-4ab8-910d-6f656ddbdf72"
    drive_id = "b!utva5lbVgkK7vSUkr4Rmyr-jw-xL7LhKkQ1vZW3b33J3_ie0A3MDS4uuoNVIHXeL"
    folder_id = "017DZBVPFPS23VFCAHI5BJ32WSP2OXVFSD"

    r = client.download_folder_contents(site_id, drive_id, folder_id, "")
    print(r)


from setuptools import setup

setup(
      name="otcfinutils",
      version="0.0.24.33",
      description="Useful functions to interact with dataverse and sharepoint",
      packages=["OTCFinUtils"],
      author="Petar Kasapinov, Shomoos Aldujaily",
      author_email="pkasapinov@otcfin.com, saldujaily@otcfin.com",
      zip_safe=False,
      install_requires=[
            "msal",
            "python-dotenv",
            "pandas",
            "azure-identity",
            "azure-keyvault-secrets",
            "openpyxl",
      ],
)
